# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import hashlib
import logging
import ssl
import sys
import time
from io import BytesIO
import certifi
import httpx
from platformdirs import user_cache_path
from .cache import _Cache

log = logging.getLogger(__name__)

# TODO: cleanup debug messages!
class _HttpClient:
    """Internal HTTP client with caching.

    Wrapper around httpx.Client with SSL context creation and 
    XDG-compliant caching.
    """
    APP_NAME = "getfactormodels"
    APP_AUTHOR = "x512"

    def __init__(self, timeout: float | int = 15.0,
                 cache_dir: str | None = None, # None by default!
                 default_cache_ttl: int = 86400):
        """Initialize the internal client.

        Args:
            timeout (str | float): max time to wait for a network response.
            cache_dir (str, optional): Path to store cached files. Defaults to
              standard user cache directory (~/.cache/getfactormodels).
            default_cache_ttl (int): cache ttl in seconds (default: 86400, one day)
        """
        self.timeout = timeout
        self.default_cache_ttl = default_cache_ttl
        self._client = None

        # XDG path
        if cache_dir is None:
            _cache_path = user_cache_path(appname=self.APP_NAME, 
                                          appauthor=self.APP_AUTHOR, 
                                          ensure_exists=True)
            self.cache_dir = str(_cache_path.resolve())

        else: # user explicitly passed a path, use it
            self.cache_dir = cache_dir
        
        self.cache = _Cache(self.cache_dir, default_timeout=default_cache_ttl)


    def __enter__(self):
        if self._client is None:

            ssl_context = ssl.create_default_context(cafile=certifi.where())
            self._client = httpx.Client(
                verify=ssl_context,
                timeout=self.timeout,
                follow_redirects=True,
                max_redirects=3,
            )
            msg = f"http_client ENTER: (client: {self._client.__class__.__name__})"
            log.debug(msg)
        return self


    def close(self) -> None:
        if self._client is not None:
            msg = f'closing {self._client.__class__.__name__}'
            log.debug(msg)  # No print in log messages, ruff
            self._client.close()
            self._client = None

        msg =f'closing {self.cache.__class__.__name__}'
        log.debug(msg)
        self.cache.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        msg = f"http_client EXIT: (client: {self._client.__class__.__name__})"
        log.debug(msg)


    def _generate_cache_key(self, url: str) -> str:
        """Generate a cache key for the URL."""
        return hashlib.sha256(url.encode('utf-8')).hexdigest()


    def download(self, url: str, model_name: str = "Data", cache_ttl: int | None = None) -> bytes:
        """Downloads content, automatically choosing between standard GET and streaming with progress."""
        key, data, expired = self._check_for_update(url)
        if not expired:
            return data

        if self._client is None:
            raise ClientNotOpenError("HttpClient is not open. Use within a 'with' block.")

        #try:
        # stream request here to get headers without downloading body yet
        with self._client.stream("GET", url) as resp:
            resp.raise_for_status()
            
            total = int(resp.headers.get("Content-Length", 0))
            # progress bar if over 1MB
            if total > 1_048_576:
                log.debug(f"File size: {total} bytes. Streaming...")
                new_data = self._progress_bar(resp, model_name)
            else:
                # Small file: just read it directly from the existing stream
                new_data = resp.read()

        #except httpx.HTTPStatusError as e:
        #    raise NetworkError(f"Server returned {e.response.status_code} for {url}") from e
        #except httpx.RequestError as e:
        #    raise NetworkError(f"Connection failed for {url}") from e

        ttl = cache_ttl or self.default_cache_ttl
        meta = {
            "etag": resp.headers.get("ETag"),
            "last_modified": resp.headers.get("Last-Modified"),
            "expires_at": time.time() + ttl,
        }
        
        self.cache.set(key, new_data, metadata=meta)
        return new_data

    def stream(self, url: str, cache_ttl: int, model_name="Model") -> bytes:
        """Wrapper around Httpx's stream.

        - Uses the _progress_bar helper. 
        - This is used by AQR models.
        """
        key, data, expired = self._check_for_update(url)
        if not expired: return data

        with self._client.stream("GET", url) as resp:
            resp.raise_for_status()
            new_data = self._progress_bar(resp, model_name)

            ttl = cache_ttl or self.default_cache_ttl
            meta = {
                "etag": resp.headers.get("ETag"),
                "last_modified": resp.headers.get("Last-Modified"),
                "expires_at": time.time() + ttl,
            }
            
            self.cache.set(key, new_data, metadata=meta)
            return new_data


    def _progress_bar(self, response, model_name: str = "Model") -> bytes:
        """A progress bar for downloads."""
        buffer = BytesIO()
        total = int(response.headers.get("Content-Length", 0))
        downloaded = 0
        start_time = time.time()
        label = f"({model_name}) Downloading data"

        chunk_size = 128 * 1024
        for chunk in response.iter_bytes(chunk_size=chunk_size):
            buffer.write(chunk)
            downloaded += len(chunk)

            if total > 0:
                percent = (downloaded / total) * 100
                elapsed = time.time() - start_time
                speed = (downloaded / 1024) / elapsed if elapsed > 0 else 0

                bar = ('#' * int(percent // 5)).ljust(20, '.')

                sys.stderr.write(f"\r{label}: [{bar}] {percent:3.0f}% ({speed:3.2f} kb/s) ")
                sys.stderr.flush()

        sys.stderr.write("\n")
        return buffer.getvalue() 
    

    # New
    def _get_metadata(self, url: str) -> dict:
        try:
            resp = self._client.head(url, timeout=5.0)
            resp.raise_for_status() 
            
            return {
                "etag": resp.headers.get("ETag"),
                "last_modified": resp.headers.get("Last-Modified"),
            }
        except Exception as e:
            # If fail (405/404, timeout), return empty, so can just do a normal download 
            log.debug(f"Err?: Unable to get metadata from {url}: {e}")
            return {}

    # New
    def _refresh_ttl(self, key, data, meta):
        """Helper to update cache's ttl without re-downloading."""
        meta["expires_at"] = time.time() + self.default_cache_ttl
        self.cache.set(key, data, metadata=meta)


    def _check_for_update(self, url: str) -> tuple[str, bytes | None, bool]:
        cache_key = self._generate_cache_key(url)
        
        cached_data, cached_meta = self.cache.get(cache_key)
        if not cached_data or not cached_meta:
            return cache_key, None, True # expired = True

        expires_at = cached_meta.get("expires_at", 0)
        if time.time() < expires_at:
            log.debug(f"CACHE HIT: {url[:30]}... ({int(expires_at - time.time())}s remaining)")
            return cache_key, cached_data, False # expired = False

        log.debug("CACHE STALE: Checking server...")
        remote_meta = self._get_metadata(url)
        
        if remote_meta.get("etag") and remote_meta.get("etag") == cached_meta.get("etag"):
            log.debug("SYNC: ETag match.")
            self._refresh_ttl(cache_key, cached_data, cached_meta)
            return cache_key, cached_data, False

        if remote_meta.get("last_modified") and remote_meta.get("last_modified") == cached_meta.get("last_modified"):
            log.debug("SYNC: Date match.")
            self._refresh_ttl(cache_key, cached_data, cached_meta)
            return cache_key, cached_data, False

        log.debug("CACHE EXPIRED: Metadata mismatch or unavailable.")
        return cache_key, None, True

    def check_connection(self, url: str):
        """Simple url ping. Boolean."""
        check_timeout = 4.0

        if self._client is None:
            raise ClientNotOpenError("HttpClient is not open. Use in a `with` block.")

        try:
            msg = f"Attempting HEAD: {url}..."
            log.info(msg)

            response = self._client.head(url, timeout=check_timeout)

            if response.is_success:
                msg = f"URL:{url}\nstatus: {response.status_code}"
                log.info(msg)
                return True

            msg = "Falling back to GET..."
            log.info(msg)

            response = self._client.get(url, timeout=check_timeout)

            if response.is_success:
                return True

            msg = "Couldn't establish connection."
            log.info(msg)
        
        except httpx.RequestError: 
            return False

        return False

    # TODO: user needs to acces this. force, or clear cache?
    def _clear_cache(self) -> None:
        self.cache.clear()
        log.debug("CACHE: cleared by HttpClient")

# TODO: Exception handling...
class ClientNotOpenError(Exception):
    """Raised when HttpClient is used outside of a 'with' block."""
    pass



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import io
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.csv as pv
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.arrow_utils import scale_to_decimal
from getfactormodels.utils.date_utils import (
    offset_period_eom,
    parse_quarterly_dates,
)


class QFactors(FactorModel): 
    """Download and process q-factor data from global-q.org.

    Args:
        frequency (str): the frequency of the data. d, m, y, q, w, w2w.
        start_date (str, optional): start date, YYYY-MM-DD.
        end_date (str, optional): end date, YYYY-MM-DD.
        classic (bool, optional): returns original 4-factor model.
        output_file (str, optional): Path to save the data automatically.
        cache_ttl (int, optional): cache TTL in seconds.

    References:
    - Hou, Kewei, Haitao Mo, Chen Xue, and Lu Zhang, 2021, An augmented 
      q-factor model with expected growth, Review of Finance 25 (1), 
      1-41.
    - Hou, Kewei, Chen Xue, and Lu Zhang, 2015, Digesting anomalies: An 
      investment approach, Review of Financial Studies 28 (3), 650-705.

    """
    @property
    def _frequencies(self) -> list[str]:
        return ["d", "w", "w2w", "m", "q", "y"]

    def __init__(self, *, classic: bool = False, **kwargs) -> None:
        """Initialize the QFactors model."""
        self.classic = classic
        super().__init__(**kwargs)
    
    @property
    def _precision(self) -> int: return 6

    @property
    def schema(self) -> pa.Schema:
        factors = [
            ("R_F", pa.float64()),
            ("R_MKT", pa.float64()),
            ("R_ME", pa.float64()),
            ("R_IA", pa.float64()),
            ("R_ROE", pa.float64()),
            ("R_EG", pa.float64()),
        ]

        if self.frequency in ["m", "q"]:
            #force 'period' here
            time_cols = [("year", pa.string()), ("period", pa.string())]
        elif self.frequency == "y":
            time_cols = [("year", pa.string())]
        else: # d/w/w2w, force 'date'
            time_cols = [("date", pa.string())]

        return pa.schema(time_cols + factors)


    def _get_url(self) -> str:
        file = {'m': "monthly", 
                "d": "daily",
                "q": "quarterly", 
                "w": "weekly",
                "w2w": "weekly_w2w",
                "y": "annual",
                }.get(self.frequency)

        url = 'https://global-q.org/uploads/1/2/2/6/122679606'
        url += f'/q5_factors_{file}_2024.csv'
        return url


    def _format_date_column(self, table: pa.Table) -> pa.Table:
        freq = self.frequency
        cols = table.column_names

        if freq == 'q':
            return parse_quarterly_dates(table=table)
        
        if "year" in cols and "period" in cols:
            yr = table.column("year").cast(pa.string())
            month = pc.utf8_lpad(table.column("period").cast(pa.string()), width=2, padding="0")
            dates = pc.binary_join_element_wise(yr, month, "")
            table = table.drop(["year", "period"]).combine_chunks()
            return table.add_column(0, "date", dates)

        if "year" in cols:
            idx = cols.index("year")
            return table.set_column(idx, "date", table.column(idx).cast(pa.string()))

        return table


    def _read(self, data: bytes) -> pa.Table:
        try:
            read_opts = pv.ReadOptions(
                column_names=self.schema.names, 
                skip_rows=1,
                block_size=1024*1024*2,
            )
            conv_opts = pv.ConvertOptions(column_types=self.schema)

            reader = pv.open_csv(
                io.BytesIO(data),
                read_options=read_opts,
                convert_options=conv_opts,
            )
            table = pa.Table.from_batches(reader)

            table = table.cast(self.schema).select(self.schema.names)
            if self.classic and "R_EG" in table.column_names:
                table = table.drop(["R_EG"])

            table = self._format_date_column(table)
            table = offset_period_eom(table, self.frequency)

            table = scale_to_decimal(table)

            renames = {"R_F": "RF_Q", "R_MKT": "Mkt-RF"}
            table = table.rename_columns([renames.get(n, n) for n in table.column_names])

            return table.combine_chunks()

        except (pa.ArrowIOError, pa.ArrowInvalid) as e:
            msg = f"{self.__class__.__name__}: reading failed: {e}"
            self.log.error(msg)
            raise ValueError(msg) from e



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import pyarrow as pa
import pyarrow.compute as pc
from getfactormodels.models.base import CompositeModel, RegionMixin
from getfactormodels.utils.utils import read_from_fred


def read_from_abs(flow_id: str, params: str, measure_val: int | str, col_name: str, client=None) -> pa.Table:
    """Sleek ABS reader for specific measures."""
    base_url = "https://data.api.abs.gov.au/rest/data"
    url = f"{base_url}/{flow_id}/{params}?detail=dataonly&format=csv"
    
    _client = client if client else _HttpClient()
    data = _client.download(url) # Uses your new streaming/get logic
    
    table = pv.read_csv(BytesIO(data))
    
    # Filter by the specific measure (e.g., 10 for ERP, 'C' for Wages)
    mask = pc.equal(table.column("MEASURE"), pa.scalar(measure_val))
    table = table.filter(mask).select(["TIME_PERIOD", "OBS_VALUE"])
    
    table = table.rename_columns(["date", col_name])
    return parse_quarterly_dates(table)

class ConditionalCAPM(CompositeModel, RegionMixin):
    """Conditional CAPM (CCAPM) of Jaganathan & Wang (1996).

    An implementations of the 'Premium-Labour', or 'Human Capital'
    CCAPM of Jaganathan and Wang (1996).

    - LBR: Smooth growth in per-capita labor income.
    - PREM: Lagged yield spread (Baa - Aaa) as a conditioning variable.

    References:
        R. Jagannathan and Z. Wang (1996). The conditional CAPM
        and the cross-section of expected returns. The Journal of
        Finance 51: 3–53.

    """
    @property 
    def _regions(self) -> list[str]: return ['usa']
    
    def __init__(self, frequency='m', **kwargs):
        super().__init__(**kwargs)
        self.frequency = frequency.lower()
        self.region = kwargs.get('region', 'usa')

    @property
    def _precision(self) -> int: return 8

    @property
    def _frequencies(self) -> list[str]: return ['m', 'q', 'y']

    @property
    def schema(self) -> pa.Schema:
        return pa.Schema([
            ('date', pa.date32()),
            ('LBR', pa.float64()),
            ('PREM', pa.float64()),
        ])

    def _construct(self, client) -> pa.Table:
        # 1. Define ABS series
        # Format: FlowID: (ColName, SDMX_Filter)
        abs_series = {
            "ABS,ERP_COMP_Q,1.0.0": ("pop", ".AUS.Q"),
            "ABS,ANA_INC,1.0.0": ("income", ".COE_WSS....Q"),
        }
        
        # 2. Get Quarterly data from ABS
        q_table = read_from_abs(abs_series, frequency='q', client=client)
        
        # 3. Scale values (Wages are usually in $M, Pop in 000s)
        # Assuming you want raw numbers for the ratio
        q_table = q_table.set_column(1, "pop", pc.multiply(q_table.column("pop"), 1000.0))
        
        # 4. Interpolate to Monthly
        # Since spread is monthly, we interpolate income/pop to match
        m_income = self._interpolate_q_to_m(q_table.select(["income"]))
        m_pop = self._interpolate_q_to_m(q_table.select(["pop"]))

    def _resample(self, table: pa.Table) -> pa.Table:
        """Downsample monthly to Q or Y by taking the last available month."""
        if self.frequency == 'm':
            return table
        months = pc.month(table.column("date"))

        if self.frequency == 'q':
            mask = pc.is_in(months, value_set=pa.array([3, 6, 9, 12]))
        elif self.frequency == 'y':
            mask = pc.equal(months, 12)
        else:
            return table

        return table.filter(mask).combine_chunks()


    def _jw_calc(self, table: pa.Table) -> pa.Table:
        """Calculate JW factors.

        - R_LBR(t): (L(t-1) + L(t-2)) / (L(t-2) + L(t-3)) - 1
        - PREM(t): Lagged spread (Baa - Aaa) at t-1
        """
        # Per Capita Labor Income (L)
        l_pc = pc.divide(table.column("income"), table.column("pop"))
        
        # Credit spread/default risk (Baa - Aaa), pct to decimal here
        spread = pc.divide(pc.subtract(table.column("baa"), table.column("aaa")), 100.0)

        n_total = table.num_rows
        # 3 lags for LBR calculation (t-1, t-2, t-3). Result starts at index 3.
        n_final = n_total - 3
        if n_final <= 0:
            raise ValueError("Insufficient data to calculate lags.")

        # Calculate R_LBR:
        # slice for the formula
        l_tm1 = l_pc.slice(2, n_final)
        l_tm2 = l_pc.slice(1, n_final)
        l_tm3 = l_pc.slice(0, n_final)

        # calc: (l_tm1 + l_tm2) / (l_tm2 + l_tm3) - 1
        r_lbr = pc.subtract(pc.divide(pc.add(l_tm1, l_tm2), pc.add(l_tm2, l_tm3)), 1.0)

        # PREM is the spread at t-1. Return period t at index 3, t-1 is index 2
        prem_lagged = spread.slice(2, n_final)

        dates = table.column("date").slice(3, n_final)

        return pa.Table.from_arrays(
            [dates, r_lbr, prem_lagged],
            names=["date", "LBR", "PREM"],
        )

    def _downsample(self, table: pa.Table) -> pa.Table:
        """Efficiently downsample monthly factors to Q or Y by filtering for EOP months."""
        if self.frequency == 'm':
            return table
        
        months = pc.month(table.column("date")) # extract the month only!
        
        if self.frequency == 'q':
            mask = pc.is_in(months, value_set=pa.array([3, 6, 9, 12]))
            table = table.filter(mask)
        
        elif self.frequency == 'y':
            # December only
            mask = pc.equal(months, 12)
            table = table.filter(mask)
            
        return table.combine_chunks()


    def _interpolate_q_to_m(self, table: pa.Table, col) -> pa.Table:
        income = table.column("income")

        q_prev = income.slice(0, len(income) - 1)
        q_curr = income.slice(1)

        # growth = (curr / prev) ^ (1/3)
        ratio = pc.divide(q_curr, q_prev)
        growth_rate = pc.power(ratio, 1/3)

        # m1 = prev / 3.0  or 1.0 if it's an index.
        m1 = pc.divide(q_prev, 3.0)
        m2 = pc.multiply(m1, growth_rate)
        m3 = pc.multiply(m2, growth_rate)

        # 3. The "Zip" Step: Interleave
        # We create a constant indices array: [0, 1, 2, 0, 1, 2, ...]
        # Since we can't use numpy, we build it via Arrow's repeat/concat logic
        num_quarters = len(m1)

        # Create the pattern [0, 1, 2]
        pattern = pa.array([0, 1, 2], type=pa.int8())

        # Repeat the pattern for every quarter and flatten
        # Note: Using a list comprehension here is okay for small arrays,
        # but for pure performance we use interleave_arrays
        indices = pa.concat_arrays([pattern] * num_quarters)

        # This "zips" m1, m2, and m3 into one long array
        monthly_income = pc.interleave_arrays([m1, m2, m3], indices)

        # return from arrays? or table?
        return pa.Table.from_arrays([monthly_income], names=["income"])
    





    # ignore this ------------------------------------------------------------------------
    def _abs_to_table(self) -> pa.Table:
        ...

    def _download_aus_erp(self, client=None) -> pa.Table:
        """Downloads population data from ABS.

        Downloads the "Population and components of change - national"
        dataset from the ABS. Quarterly data.

        Data:
        Data Item:
        """
        # ABS REST - AUS: Quarterly (Q): Estimated Resident Population (ERP)
        base_url = "https://data.api.abs.gov.au/rest/data"
        url = f"{base_url}/ABS,ERP_COMP_Q,1.0.0/.AUS.Q?detail=dataonly&format=csv"

        _client = client if client else _HttpClient()

        data = _client.stream(url, cache_ttl=20000)

        table = pv.read_csv(io.BytesIO(data))

        # ERP (MEASURE = 10)
        mask = pc.equal(table.column("MEASURE"), 10)
        pop_t = table.filter(mask)

        pop_t = pop_t.select(["TIME_PERIOD", "OBS_VALUE"])
        pop_t = pop_t.rename_columns(["date", "pop"])

        # Data's in units of 000's
        abs_pop_table = pop_t.set_column(
            1, "pop", pc.multiply(pop_table.column("pop"), 1000.0),
        )
        # parse qtrly, offset
        abs_pop_table = parse_quarterly_dates(abs_pop_table)
        table = offset_period_eom(abs_pop_table)

        table.validate()

        print(table)


    def _download_aus_wages(self, client=None) -> pa.Table:
        """Downloads wage date from ABS.

        Quarterly data.

        Data: Australian National Accounts – Income from Gross Domestic Product
        Data Item: Compensation of employees - Wages and salaries
        """
        url = "https://data.api.abs.gov.au/rest/data/ABS,ANA_INC,1.0.0/.COE_WSS....Q?detail=dataonly&format=csv"
        _client = client if client else _HttpClient()

        data = _client.stream(url, cache_ttl=20000)
        table = pv.read_csv(io.BytesIO(data))

        mask = pc.and_(
            pc.equal(raw_table.column("TSEST"), pa.scalar(20)),
            pc.equal(raw_table.column("MEASURE"), pa.scalar("C")),
        )

        table = table.filter(mask)

        # 3. Now select and rename
        table = table.select(["TIME_PERIOD", "OBS_VALUE"])
        table = table.rename_columns(["date", "income"])

        # 4. Process dates (YYYY-QX -> Date32)
        #d_str = table.column("date").cast(pa.string())
        #years = pc.utf8_slice_codeunits(d_str, start=0, stop=4)
        #q_nums = pc.utf8_slice_codeunits(d_str, start=6, stop=7)
        #m_map = pa.array(["03", "06", "09", "12"])
        #m_idx = pc.subtract(q_nums.cast(pa.int32()), 1)
        #months = pc.take(m_map, m_idx)
        #iso_dates = pc.binary_join_element_wise(pc.binary_join_element_wise(years, months, "-"), pa.scalar("01"), "-")
        
        table = parse_quarterly_dates(table)
        table = offset_period_eom(table)

        #table = table.set_column(0, "date")
        #table = table.rename_columns([0, 'date'])
        table = table.sort_by("date")

        table





#### MOVED HERE FOR A BIT

# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import io
import pyarrow as pa
import pyarrow.csv as pv
from getfactormodels.models.base import CompositeModel, RegionMixin
from getfactormodels.utils.date_utils import (
    offset_period_eom,
    parse_quarterly_dates,
)
from getfactormodels.utils.http_client import _HttpClient


# might rename PremiumLabourCCAPM
#class PremiumLabourCAPM(CompositeModel, RegionMixin):
class ConditionalCAPM(CompositeModel, RegionMixin):
    """Conditional CAPM (CCAPM) of Jaganathan & Wang (1996).

    An implementations of the 'Premium-Labour', or 'Human Capital' 
    CCAPM of Jaganathan and Wang (1996). 

    Data from FRED. Monthly, except income, 

    - LBR: Smooth growth in per-capita labor income.
    - PREM: Lagged yield spread (Baa - Aaa) as a conditioning variable.

    References:
        R. Jagannathan and Z. Wang (1996). The conditional CAPM
        and the cross-section of expected returns. The Journal of 
        Finance 51: 3–53.

    """
    _regions = ["usa", "aus"]
    def __init__(self, **kwargs):
        region = kwargs.pop('region', None)
        super().__init__(**kwargs)
        self.region = region

    @property
    def _frequencies(self) -> list[str]:
        return ['m', 'q', 'y']

    @property 
    def _precision(self) -> int:
        return 8

    @property
    def _regions(self) -> list[str]:
        return ['aus', 'usa'] #testin

    @property 
    def schema(self) -> pa.Schema:
        return pa.Schema([
            ('date', pa.date32()),
            ('LBR', pa.float64()),
            ('PREM', pa.float64()),
        ])


    def _construct(self, client) -> pa.Table:
        m_series = {
            "POPTHM": "pop",
            "BAA": "baa",
            "AAA": "aaa",
        } if self.region == 'usa' else {
            # Testing something...
            "POPTOTAUA647NWDB": "pop",
            "IR3TBB01AUM156N": "baa",
            "IRLTLT01AUM156N": "aaa",
        }
        
        m_table = read_from_fred(series=m_series, frequency='m', client=client)
        m_table = offset_period_eom(m_table, 'm')
        
        # Income data from FRED (COE) is quarterly.
        q_id = "LCWRTT01AUQ661N" if self.region == 'aus' else "COE"

        q_table = read_from_fred(series={q_id: "income"}, frequency='q', client=client)
        q_table = offset_period_eom(q_table, frequency='q').combine_chunks()

        # helper: quarterly COE to monthly 
        inc_table = self._q_wages_to_m(q_table)

        data_table = m_table.join(inc_table, keys="date", join_type="inner")
        data_table = data_table.sort_by("date").combine_chunks()

        jw_table = self._jw_calc(data_table)

        # scale PREM, not importing scale_to_decimal just for this
        prem_dec = pc.divide(jw_table.column("PREM"), 100.0)

        table = jw_table.set_column(2, "PREM", prem_dec).combine_chunks()
        return self._downsample(table)
    
        #ff = FamaFrenchFactors(model='3', frequency=self.frequency).load(client=client)
        #mkt = select_table_columns(ff.data, ['Mkt-RF', 'RF'])
        #return jw_table.join(mkt, keys="date", join_type="inner").combine_chunks()


    def _jw_calc(self, table: pa.Table) -> pa.Table:
        """J&W (1996) factor construction logic."""
        # Per Capita Labor Income (L)
        l_pc = pc.divide(table.column("income"), table.column("pop"))
        spread = pc.subtract(table.column("baa"), table.column("aaa"))
        
        n_total = table.num_rows
        n_final = n_total - 3 # We lose 3 rows due to t-3 lag
        
        if n_final <= 0:
            raise ValueError("Insufficient data rows for ConditionalCAPM lags.")

        # Slices: All must have length = n_final
        # t is index 3...t-3 is index 0.
        l_tm1 = l_pc.slice(2, n_final)
        l_tm2 = l_pc.slice(1, n_final)
        l_tm3 = l_pc.slice(0, n_final)
        
        # JW's R_LBR: (L_t-1 + L_t-2) / (L_t-2 + L_t-3) - 1
        num = pc.add(l_tm1, l_tm2)
        den = pc.add(l_tm2, l_tm3)

        r_labor = pc.subtract(pc.divide(num, den), 1.0)
        
        # PREM (t-1 spread)
        prem_lagged = spread.slice(2, n_final)

        dates = table.column("date").slice(3, n_final)

        return pa.Table.from_arrays(
            [dates, r_labor, prem_lagged],
            names=["date", "LBR", "PREM"],
        ).combine_chunks()





# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import logging
import pyarrow as pa
import pyarrow.compute as pc

"""Utilities for working with pyarrow."""

log = logging.getLogger(__name__)


def scale_to_decimal(table: pa.Table) -> pa.Table:
    """Standardize float cols to decimal (5.2 -> 0.052)."""
    for i, field in enumerate(table.schema):
        if pa.types.is_floating(field.type):
            scaled_col = pc.divide(table.column(i), 100.0)
            table = table.set_column(i, field.name, scaled_col)
    return table


def round_to_precision(table: pa.Table, precision: int) -> pa.Table:
    """Rounds all float cols in a pa.Table to precision, and RF to 4.

    Note:
    Used by all models.
    """
    # auto resolve to: precision arg, else model's ._precision or '6'.
    prec_val = precision or getattr(table, '_precision', 6)
    rf_cols = {'RF', 'R_F'} # TODO: fix this up.

    new_cols = []
    for i, field in enumerate(table.schema):
        col = table.column(i)
        if pa.types.is_floating(field.type):
            p = 4 if field.name.upper() in rf_cols else 6 if field.name.upper() == "RF_AQR" else prec_val
            col = pc.round(col, ndigits=p, round_mode='half_to_even')
        new_cols.append(col)

    return pa.Table.from_arrays(new_cols, schema=table.schema)


def rearrange_columns(table: pa.Table) -> pa.Table:
    """Standardize column orders: 'date', [FACTORS], 'RF'"""
    cols = table.column_names
    front = [c for c in ['date', 'Mkt-RF'] if c in cols]
    back = [c for c in ['RF_AQR', 'RF_Q', 'RF'] if c in cols]
    fixed = set(front + back) #fix: don't define set in the loop! Now list comp is O(1)
    mid = [c for c in cols if c not in fixed]
    return table.select(front + mid + back)


def filter_table_by_date(table: pa.Table,
                         start: str | None,
                         end: str | None) -> pa.Table:
    """Slices a table to a start-end range."""
    if start is None and end is None:
        return table

    # target the date col and its type
    date_col = 'date' if 'date' in table.column_names else table.column_names[0]
    target_type = table.schema.field(date_col).type
    expr = pc.field(date_col)

    mask = None
    if start:
        mask = (expr >= pc.scalar(start).cast(target_type))

    if end:
        end_mask = (expr <= pc.scalar(end).cast(target_type))
        mask = (mask & end_mask) if mask is not None else end_mask

    return table.filter(mask)


def _validate_columns(table: pa.Table, names: str | list[str] | None) -> list[str]:
    """Private helper: Checks if columns exist in a pyarrow Table by name.

    Returns the list of validated factors or raises if missing.
    """
    if table.num_rows == 0:
        raise RuntimeError("Table is empty.")

    if not names:
        return []

    input_list = [names] if isinstance(names, str) else list(names)

    cols = set(table.column_names)

    missing = [f for f in input_list if f not in cols]
    if missing:
        raise ValueError(f"Columns not found in Table: {missing}")

    return input_list


def select_table_columns(table: pa.Table, factors: str | list[str]) -> pa.Table:
    """Helper: constructs a table from a pa.Table by selecting cols by name.

    'date' is col 0, and if passed in 'factors', is depduplicated.
    Used by base FactorModel's __getitem__, _extract_as_table and .extract()

    Case insensitive.
    """
    req_cols = [factors] if isinstance(factors, str) else list(factors)

    name_map = {col.lower(): col for col in table.column_names}

    valid_names = [name_map.get(f.lower(), f) for f in req_cols]
    valid_cols = _validate_columns(table, valid_names)

    factor_cols = [f for f in valid_cols if f.lower() != 'date']
    if not factor_cols:
        raise ValueError("Extraction must include at least one factor.")   ### TEST THIS

    # ['date'] + [verified factors minus date]
    selection = list(dict.fromkeys(['date'] + factor_cols))

    return table.select(selection).combine_chunks()


def _format_for_preview(val, col_name, precision=6):
    """Private helper for print_table_preview output. Rounds for display converts None to NaNs."""
    if val is None or (isinstance(val, float) and val != val): 
        return "NaN"

    if isinstance(val, (float, int)) and col_name != "date":
        prec = 4 if col_name.upper() in {'RF', 'RF_AQR'} else precision
        return f"{val:.{prec}f}"

    return str(val)


def print_table_preview(table, n_rows=4) -> str | None:
    """Prints a pa.Table like a (simplified) pd.DataFrame.

    Used by the cli, and base class __repr__ (TODO).

    Notes:
        - Hardcoded 'date' and '[...]' for the index col and gap. 
        - Head and tails (n_rows) 4 by default. 
        - Table size is included in the footer (might change to a nice header)
        - Will print whole table if total is less than (n_rows * 2) + 5. 
    """
    total_rows = table.num_rows
    if table.num_rows == 0:
        log.warning("Empty Table")  # no sys
        return None

    # Get the model's own _precision property!
    precision = getattr(table, '_precision', 6)
    columns = table.column_names

    # if there's a gap, then the row is None. (Below, if None, prints "[...]")
    is_gap = total_rows <= (n_rows * 2) + 5   # +5 to prevent hiding just a few cols

    # slice a head and tail
    head = table.slice(0, n_rows).to_pylist() if not is_gap else table.to_pylist()
    tail = table.slice(total_rows - n_rows, n_rows).to_pylist() if not is_gap else []

    # calc width of cols
    col_widths = {
        col: max(
            len(col),
            max((len(_format_for_preview(r[col], col, precision)) for r in head + tail), default=0),
        ) + 2 
        for col in columns
    }

    # construct header (collect into list, no printing to stderr)
    output = []
    header_row = "".join(c.rjust(col_widths[c]) for c in columns[1:])
    output.append(f"\t    {header_row}")
    output.append("date")

    # print (head, gap, tail) rows. TODO: nice header, __str__ in base
    display_rows = head + ([None] if tail else []) + tail
    for row in display_rows:
        if row is None:    # from above, row's None, because is_gap
            output.append("  [...] ")   ### TEST THIS
            continue

        # left align dates, other cols right
        date_str = _format_for_preview(row['date'], 'date', precision).ljust(col_widths['date'])
        factors_str = "".join(_format_for_preview(row[c], c, precision).rjust(col_widths[c]) for c in columns[1:])
        output.append(f"{date_str}{factors_str}")

    buf_size = table.get_total_buffer_size()
    output.append(f"\n[{total_rows} rows x {len(columns)} columns, {buf_size / 1024:.1f} kb]")
    # change: return the str, not print
    return "\n".join(output)



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import io
import re
import pyarrow as pa
import pyarrow.csv as pv
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.date_utils import offset_period_eom


class LiquidityFactors(FactorModel):
    """Liquidity factors of Pastor-Stambaugh (2003).

    Download the Pastor-Stambaugh Liquidity Factors. Only available 
    in monthly.

    Args:
        frequency (str): The data frequency, 'm'.
        start_date (str, optional): The start date YYYY-MM-DD.
        end_date (str, optional): The end date YYYY-MM-DD.
        output_file (str, optional): file path to save to.
          Supports csv, pkl.
        cache_ttl (int, optional): download time-to-live in secs 
        (default: 86400).

    References:
    - L. Pastor and R. Stambaugh, ‘Liquidity Risk and Expected Stock 
      Returns’, Journal of Political Economy, vol. 111, no. 3, pp. 
      642–685, 2003.

    ---

    Note:
        the leading 65 values of TRADED_LIQ are NaNs.
    """
    @property
    def _frequencies(self) -> list[str]: return ["m"]

    @property
    def _precision(self) -> int: return 8

    @property
    def schema(self) -> pa.Schema:
        return pa.schema([
            ('Month', pa.int64()),
            ('Agg Liq.', pa.float64()),
            ('Innov Liq (eq8)', pa.float64()),
            ('Traded Liq (LIQ_V)', pa.float64()),
        ])

    def _get_url(self) -> str:
        return 'https://finance.wharton.upenn.edu/~stambaug/liq_data_1962_2024.txt'

    def _read(self, data: bytes) -> pa.Table:
        try:
            _text = data.decode('utf-8')
            _lines = [
                re.sub(r'\s+', '\t', line.strip()) 
                for line in _text.splitlines() 
                if line.strip() and not line.startswith('%')
            ]

            _data = '\n'.join(_lines).encode('utf-8')

            # open stream reader not read whole thing
            reader = pv.open_csv(
                io.BytesIO(_data),
                read_options=pv.ReadOptions(
                    column_names=self.schema.names,
                    block_size=1024*1024*2,
                ),
                parse_options=pv.ParseOptions(delimiter='\t'),
                convert_options=pv.ConvertOptions(
                    column_types=self.schema,
                    null_values=["-99", "-99.0", "-99.000000"],
                ),
            )

            # Batch processing
            batches = []
            for batch in reader:
                batches.append(batch)

            # .from_batches!!!
            table = pa.Table.from_batches(batches)
            table = offset_period_eom(table, self.frequency)
            #table.validate()
            return table.rename_columns(['date', 'AGG_LIQ', 
                                         'INNOV_LIQ', 'TRADED_LIQ'])   #TODO: fix, not enforcing schema...

        except (pa.ArrowIOError, pa.ArrowInvalid) as e:
                msg = f"{self.__class__.__name__}: reading failed: {e}"
                self.log.error(msg)
                raise ValueError(msg) from e



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import calendar
import logging
import re
import sys
from datetime import date, datetime
from typing import Any
import pyarrow as pa
import pyarrow.compute as pc

log = logging.getLogger(__name__)

def offset_period_eom(table: pa.Table, frequency: str) -> pa.Table:
    """Utility to offset a pa.Table's date column to the last day of its period."""
    if table.num_columns == 0:
        return table
        
    orig_name = table.column_names[0]
    freq = frequency.lower()

    # remove non digits 
    raw = pc.replace_substring_regex(table.column(0).cast(pa.string()), r"\D", "")
    lengths = pc.utf8_length(raw)

    # Constructs a valid date str. If 8 chars, keep. If 4 or 6 append 0101 or 01. 
    clean = pc.case_when(
        pc.make_struct(pc.equal(lengths, 8), pc.equal(lengths, 6), pc.equal(lengths, 4)),
        raw,                                                    # 8
        pc.binary_join_element_wise(raw, pa.scalar("01"), ""),  # 6
        pc.binary_join_element_wise(raw, pa.scalar("0101"), ""), # 4
    )
    dates = pc.strptime(clean, format="%Y%m%d", unit="s")

    if freq in ['m', 'q', 'y']:
        # Map the frequency to ceil_temporal's unit. 
        unit_map = {'m': 'month', 'q': 'quarter', 'y': 'year'}
        unit = unit_map[freq]

        # Add a day to make sure we're strictly inside the period.
        _one_day = pa.scalar(86400000, type=pa.duration('ms'))
        extra_day = pc.add(dates, _one_day)

        # From the start of the next period, subtract one day to get EOM/Q/Y
        next_period_start = pc.ceil_temporal(extra_day, unit=unit)
        final_dates = pc.subtract(next_period_start, _one_day)
    else: # 'd' or unknown, just return the parsed dates.
        final_dates = dates

    return table.set_column(0, orig_name, final_dates.cast(pa.date32()))


# ALL retreived quarterly dates: Q factors, ABS, ICR, fred
def parse_quarterly_dates(table: pa.Table) -> pa.Table:
    """Converts quarterly dates to a datetime str.

    Converts 1 column, "yyyyq", or two columns, "year" and "period" to a single 
    'date' column, of iso datestr yyyy-mm-dd, set to the end of the period.

    - Used by ICRFactors() and QFactors().
    """
    if table.num_rows == 0:
        return table

    # q-factors: 2 cols - 'year' and 'period'
    if "year" in table.column_names and "period" in table.column_names:
        year = table.column("year").cast(pa.string())
        qtr = table.column("period").cast(pa.int32())
        table = table.drop(["year", "period"])
    else: 
        dates = table.column(0).cast(pa.string())
        year = pc.utf8_slice_codeunits(dates, 0, 4)
        qtr = pc.utf8_slice_codeunits(dates, start=-1).cast(pa.int32())
        table = table.remove_column(0)
    m_int = pc.multiply(qtr, 3).cast(pa.string())
    month = pc.utf8_lpad(m_int, width=2, padding="0")

    days = pa.array(["01"] * table.num_rows, type=pa.string())
    datestr = pc.binary_join_element_wise(year, month, days, "-")
    
    try:
        date_col = datestr.cast(pa.date32())
        # Re-insert as the first column
        return table.add_column(0, "date", date_col).combine_chunks()
    except pa.ArrowInvalid as e:
        raise ValueError(f"Failed to parse quarterly dates: {e}")


def validate_date_range(start: str | None, end: str | None) -> tuple[str | None, str | None]:
    """Validates start and end date range. Sets end to today if in the future.
    
    If start is in the future, err.
    """
    today = date.today().isoformat()

    if start and end and start > end:
        raise ValueError(f"start_date ({start}) is after end_date ({end}).")

    if start and start > today:
        raise ValueError(f"Start date is in the future. {start} > {today}.") 
    
    if end and end > today:
        print(f"WARNING: End date is in the future: {end}. Setting to today: {today}.", file=sys.stderr)
        end = today

    return start, end


def _roll_to_eom(dt: datetime) -> str:
    """Roll a datetime to the last day of its month.
    
    Used for user input only.
    """
    last_day = calendar.monthrange(dt.year, dt.month)[1]
    return dt.replace(day=last_day).strftime("%Y-%m-%d")


def _validate_input_date(date_input: Any, *, is_end: bool = False) -> str | None:
    if date_input is None:
        return None

    raw_str = str(date_input).strip()

    if '-' in raw_str:
        parts = raw_str.split('-')
        year = parts[0]
        month = parts[1].zfill(2) # don't assume a day exists
        
        if len(parts) >= 3:
            day = parts[2].zfill(2)
            raw_str = f"{year}-{month}-{day}"
        else:
            raw_str = f"{year}-{month}"

    clean_str = re.sub(r'\D', '', raw_str) 
    try:
        if len(clean_str) == 4:  # YYYY
            dt = datetime(int(clean_str), 12, 31) if is_end else datetime(int(clean_str), 1, 1)
        elif len(clean_str) == 6: # YYYYMM
            dt = datetime.strptime(clean_str, "%Y%m")
            if is_end:
                dt = dt.replace(day=calendar.monthrange(dt.year, dt.month)[1])
        elif len(clean_str) == 8: # YYYYMMDD
            dt = datetime.strptime(clean_str, "%Y%m%d")
        else:
            try:
                dt = datetime.fromisoformat(raw_str.replace('/', '-'))
            except ValueError:
                dt = datetime.strptime(raw_str, "%Y-%m")
                if is_end:
                    dt = dt.replace(day=calendar.monthrange(dt.year, dt.month)[1])

        return dt.strftime("%Y-%m-%d")

    except Exception:
        raise ValueError(f"Invalid date: '{date_input}'. Use YYYY, YYYY-MM, or YYYY-MM-DD.")

# test:
# yyyy, yyyy-mm, yyyymm, yyyymmdd end dates 
# That they return as intended, and yyyy-mm isn't filling in today's day.
# isn't reading YYYYMM as YYMMDD (201204 becoming 2020-12-04)
# isn't filling in today's month or current year.
# IS TESTED WITH ALL RETURNED VALUES IN SOURCE DATA.
#TODO: style warnings! Consistent logging!



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any
import pyarrow as pa
from getfactormodels.utils.arrow_utils import (
    filter_table_by_date,
    print_table_preview,
    rearrange_columns,
    round_to_precision,
    select_table_columns,
)
from getfactormodels.utils.date_utils import (
    _validate_input_date,
    validate_date_range,
)
from getfactormodels.utils.http_client import _HttpClient
from getfactormodels.utils.utils import _save_to_file

"""An abstract base class for model implementations.

- FactorModel: abstract base class. Provides common data handling, 
  caching, and date-filtering logic implemented by all factor models.
- RegionMixin: A mixin for models supporting international data.
- CompositeModel: base class for models built from other models.
- ModelCollection: for returning multiple models in a combined table.
"""
logger = logging.getLogger(__name__)


class FactorModel(ABC):
    """Abstract Base Class used by all factor model implementations."""
    def __init__(self, frequency: str | None = 'm',
                 start_date: str | None  = None,
                 end_date: str | None = None,
                 output_file: str | None = None,
                 cache_ttl: int = 86400,
                 **kwargs: Any):
        """Initialize the factor model instance.

        Args:
            frequency (str): the frequency of the data. Default: 'm'.
            start_date (str, opt): The start date YYYY[-MM-DD]
            end_date (str, opt): The end date YYYY[-MM-DD]
            output_file (str): optional path to save data to.
            cache_ttl (int): Cache time-to-live in seconds. Default: 86400.
            **kwargs: some models have additional params.
        """
        self.log = logger

        self._data: pa.Table | None = None
        self._start_date = None
        self._end_date = None
        self._frequency = None

        self.frequency = frequency 
        self.start_date = start_date 
        self.end_date = end_date
        self.output_file = output_file
        self.cache_ttl = cache_ttl

        self.copyright: str = ""  # NEW, TEST. fix: Carhart erroring with FF with copyright
        self._selected_factors: list[str] | None = None  # for eg drop/extract  # changing
        super().__init__()

    def __len__(self) -> int:
        """Length of the pa.Table after filtering."""
        return len(self.data)

    def __str__(self) -> str:
        if self.data is None:
            return self.__repr__()
        
       # table = self.data

        if isinstance(self, RegionMixin):
            region_label = f" ({self.region})"
        else:
            region_label = ""

        header = f"{self.__class__.__name__}{region_label}\n"
        return header + print_table_preview(self.data)

    def __repr__(self) -> str:
        params = []
        attrs = [
            'model', 'frequency', 'region', 'start_date', 'end_date',
            'country', 'classic', 'output_file',
        ]

        for attr in attrs:
            val = getattr(self, attr, None)

            if val is not None and val is not False and val != "":
                # format the str with quotes
                repr_val = f"'{val}'" if isinstance(val, str) else val
                params.append(f"{attr}={repr_val}")

        return f"{self.__class__.__name__}({', '.join(params)})"

    def __getitem__(self, key: str | list[str]) -> pa.Table:
        """Returns a pa.Table of date + selected factors."""
        return select_table_columns(self.data, key)

    def _repr_html_(self) -> str:
        """HTML repr for nice IPython/Jupyter outputs."""
        style = (
            "font-family: monospace; "
                "font-size: 0.9em; "
                "line-height: 1.4; "
                "background-color: transparent; "
                "white-space: pre; "
                "overflow-x: auto; "
                "display: block;"
        )

        # str(self) calls get_table_preview, wrap in <pre>
        return f'<pre style="{style}">{str(self)}</pre>'

    @property
    def start_date(self) -> str | None:
        return self._start_date
    @start_date.setter
    def start_date(self, value: Any):
        valid = _validate_input_date(value, is_end=False)
        self._start_date, self._end_date = validate_date_range(valid, self._end_date)

    @property
    def end_date(self) -> str | None:
        return self._end_date
    @end_date.setter
    def end_date(self, value: Any):
        valid = _validate_input_date(value, is_end=True)
        self._start_date, self._end_date = validate_date_range(self._start_date, valid)

    @property
    def frequency(self) -> str | None: 
        return self._frequency
    @frequency.setter
    def frequency(self, value: str | None):
        if value is None:
            self._frequency = None
            return

        val = value.lower()
        if val not in self._frequencies:
            raise ValueError(f"Invalid '{val}'. Options: {self._frequencies}") 

        if val != self._frequency:
            if self._frequency is not None:
                msg = f"Freq. changed from {self._frequency} to {val}." 
                self.log.info(msg)
            self._frequency = val
            self._data = None


    @property
    def data(self) -> pa.Table:
        """Returns a pa.Table with requested data (sliced)."""
        if self._data is None:
            self.load()   # Don't know if good idea
            
        table = self._data

        # drop/extract handled here.
        if self._selected_factors is not None:
            selection = list(dict.fromkeys(['date'] + self._selected_factors))
            table = table.select(selection)

        return filter_table_by_date(table, self._start_date, self._end_date)


    @property
    def shape(self) -> tuple[int, int]:
        """(rows, columns), like Pandas/Numpy."""
        return self.data.shape


    def extract(self, factor: str | list[str]) -> "FactorModel":   #Self
        """Select specific factors from the model. Str or list[str]."""
        table = select_table_columns(self.load(), factor)
        self._selected_factors = [f for f in table.column_names if f != 'date']
        return self


    def drop(self, factor: str | list[str]) -> "FactorModel": #Self
        """Remove specific factors from the model. Str or list[str]."""
        t_cols = self.load().column_names
        to_drop = {f.lower() for f in ([factor] if isinstance(factor, str) else factor)}

        # select: cols (lowercase) not in the to_drop set
        selection = [c for c in t_cols if c.lower() not in to_drop and c != 'date']

        if not selection:
            raise ValueError("Cannot drop all factors from the model.")

        self._selected_factors = selection
        return self


    def to_file(self, filepath: str | Path | None = None) -> None:
        """Save data to a file.

        Supports .parquet, .ipc, .feather, .csv, .txt, .pkl, .md
        
        Args:
            filepath (str | Path, optional): filepath to save data to.
        """
        target = filepath if filepath else self.output_file
        if not target:
            self.log.error("No filepath provided and no default output_file set.")
            return

        # self.data does get_table, selection, and date slicing.
        table = self.data

        if table.num_rows == 0:
            self.log.warning("No data available to save.")
            return

        _save_to_file(table, target, model_instance=self)


    def to_pandas(self) -> "pd.DataFrame":
        """Convert model to a pandas DataFrame.

        - Triggers the download if not loaded.
        """
        try:
            import pandas as pd  # check for if user has pd?
            df = self.data.to_pandas()
            if "date" in df.columns:
                df = df.set_index("date")
            return df
        except ImportError:
            raise ImportError("Requires Pandas. Try `pip install pandas`") from None


    def to_polars(self) -> "pl.DataFrame":
        """Convert model to a polars DataFrame.

        - Wrapper around Polars' `from_arrow()`. 
        - Triggers the download if not loaded.
        """
        try:
            import polars as pl
            return pl.from_arrow(self.data)
        except ImportError:
            raise ImportError("Requires Polars. Try `pip install polars`") from None
    # maybe
    #def preview(self, n: int = 4):
    #    """Prints the formatted table preview."""
    #    print(print_table_preview(self.data, n_rows=n))

    # RENAME: load, was _get_table
    def load(self, client: _HttpClient | None = None) -> pa.Table:
        """Trigger download or construction."""
        if self._data is not None:
            return self._data

        if hasattr(self, '_construct'):
            if client is None:
                with _HttpClient() as shared_client:
                    table = self._construct(shared_client)
            else:
                table = self._construct(client)
        else:
            raw_bytes = self._download(client=client)
            table = self._read(raw_bytes)

        # move this out probably ---------------------
        if "date" in table.column_names:
            table = table.sort_by([("date", "ascending")])

        table = round_to_precision(table, self._precision)

        # CompositeModel: if drop_null=True, drop
        if getattr(self, 'drop_null', False):
            table = table.drop_null()

        table.validate(full=True)
        self._data = rearrange_columns(table=table).combine_chunks()
        # -------------------------------------------
        return self


    def _download(self, client: _HttpClient | None = None) -> bytes | dict[str, bytes]:
        urls = self._get_url()
        self.log.info(f"Downloading from: {urls}")
        def _download_method(client: _HttpClient, urls: str | dict):
            if isinstance(urls, str):
                # A client was given. Using it.
                return client.download(urls, self.cache_ttl)
            return {k: client.download(v, self.cache_ttl) for k, v in urls.items()}
        try:
            if client:
                return _download_method(client, urls)
            with _HttpClient() as new_client:
                return _download_method(new_client, urls)
        except Exception as e:
            self.log.error(f"Download failed: {e}")
            raise RuntimeError(f"Could not retrieve data for {self.__class__.__name__}") from e


    # might move to utils
    def __dataframe__(self, *, nan_as_null: bool = False):
        """Dataframe interchange protocol support.

        Casts date32 to ns.

        Args:
            nan_as_null: converts NaN to null

        Examples:
            model = FamaFrenchFactors(model='3')
            df = model.to_pandas()

            import polars as pl
            df = pl.from_arrow(model.data)

            import pandas as pd
            df = pd.api.interchange.from_dataframe(model.data)
        """
        table = self.data

        if table.column_names[0] == "date":
            col_type = table.schema.field(0).type

        if pa.types.is_date(col_type):
            date_ns = table.column(0).cast(pa.timestamp("ns")).combine_chunks()
            table = table.set_column(0, "date", date_ns)

        return table.combine_chunks().__dataframe__(nan_as_null=nan_as_null, allow_copy=True)


    @property
    def _precision(self) -> int: return 8

    @property
    @abstractmethod
    def schema(self) -> pa.Schema:
        pass

    @property
    @abstractmethod
    def _frequencies(self) -> list[str]:
        pass

    @abstractmethod 
    def _get_url(self) -> str | dict[str, str]:
        """Build the unique data source URL."""
        pass

    @abstractmethod
    def _read(self, data: bytes) -> pa.Table:
        """Read bytes into a pa.Table."""
        pass


class CompositeModel(FactorModel):
    """Base for models constructed from other models."""
    def __init__(self, frequency: str = 'm', *, drop_null: bool = True, **kwargs):
        self.drop_null = drop_null
        super().__init__(frequency=frequency, **kwargs)

    @abstractmethod
    def _construct(self, client: _HttpClient) -> pa.Table:
        pass
    
    def _get_url(self) -> str:
        raise NotImplementedError("CompositeModel: no remote source.")
    
    def _read(self, data: bytes) -> pa.Table:
        raise NotImplementedError("CompositeModel: _read called on a composite models.")


class ModelCollection(CompositeModel):
    """A ModelCollection holds multiple models of the same frequencies."""
    def __init__(self, model_keys: list[str], **kwargs):
        from getfactormodels.main import get_factors

        self.model_keys = model_keys
        self.instances = [get_factors(m, **kwargs) for m in model_keys]

        regional_inst = [i for i in self.instances if isinstance(i, RegionMixin)]
        
        if regional_inst and len(regional_inst) < len(self.instances):
            req_region = kwargs.get('region')
            if req_region and req_region.lower() not in ['us', 'usa']: #allow USA through, as all models are US models. 
                raise ValueError("Cannot combine regional models with non-regional models for non-US regions.")
        super().__init__(**kwargs)


    def __str__(self) -> str:
        if self._data is None:
            return self.__repr__()

        region_label = ""
        regional_inst = next((i for i in self.instances if isinstance(i, RegionMixin)), None)
        if regional_inst:
            region_label = f" ({regional_inst.region})"

        models_str = ", ".join(f"'{k}'" for k in self.model_keys)
        header = f"{self.__class__.__name__}: {models_str} {region_label}\n"
        return header + print_table_preview(self.data)

    @property
    def _frequencies(self) -> list[str]:
        # the intersection of frequencies btw models
        freqs = set(self.instances[0]._frequencies)
        for inst in self.instances[1:]:
            freqs &= set(inst._frequencies)
        return list(freqs)

    @property
    def schema(self) -> pa.Schema:
        seen = {'date'}
        fields = [pa.field('date', pa.date32())]
        for inst in self.instances:
            for field in inst.schema:
                if field.name not in seen:
                    fields.append(field)
                    seen.add(field.name)
        return pa.schema(fields)
    

    def _construct(self, client: _HttpClient) -> pa.Table:
        tables = []
        for inst in self.instances:
            inst.load(client=client)
            table = inst.data
            table = table.set_column(0, "date", table.column("date").cast(pa.date32()))
            tables.append(table)

        # start with the first table
        final_table = tables[0]
        
        for i, next_table in enumerate(tables[1:], 1):
            # finding (probable) duplicate columns (not 'date') TODO: rename some factors...
            overlaps = set(final_table.column_names) & set(next_table.column_names)
            overlaps.discard('date')
            
            keep_cols = [c for c in next_table.column_names 
                                  if c not in overlaps and c != 'date']

            if keep_cols:
                cols_to_join = ['date'] + keep_cols
                table_to_join = next_table.select(cols_to_join)

                final_table = final_table.join(
                    table_to_join, 
                    keys="date", 
                    join_type="left outer",
                )
            
        return final_table.sort_by([("date", "ascending")])


class RegionMixin:
    """Mixin for models that support international regions/countries."""
    #TODO: map properly!
    _aliases = {
        'uk': 'gbr', 'ger': 'deu', 'sg': 'sgp',
        'hk': 'hkg', 'nl': 'nld', 'fr': 'fra',

        # FF and AQR (not VME) - shared countries
        'usa': 'us', 'us': 'usa',
        'jpn': 'japan', 'japan': 'jpn',

        # FF and AQR (not VME) - shared regions : aggregate portfolios
        'ex-us': 'global ex usa',
        'north-america': 'north america', 'north america': 'north-america',

        # FF regions : AQR Aggregate Portfolios. 
        # Note: AQR's 'pacific' includes japan.
        #for VME
        'global_stocks': 'all_equities',
        
        # Aus
        'australia': 'aus', 'au': 'aus',
    }

    @property
    @abstractmethod
    def _regions(self) -> list[str]:
        pass

    @property
    def region(self) -> str:
        if hasattr(self, "_region") and self._region is not None:
            return self._region
        if not self._regions:
            return ""

        # set default: usa, else first region if no usa in model's regions... 
        if 'us' in self._regions:
            return 'us'
        if 'usa' in self._regions:
            return 'usa'
        return self._regions[0]
            
    @region.setter
    def region(self, value: str | None):
        if value is None:
            self._region = None
            return

        val = str(value).strip().lower()
        # Try a direct match, then alias, else keep (to fail)
        resolved = val if val in self._regions else self._aliases.get(val, val)
        
        if resolved not in self._regions:
            raise ValueError(f"Region '{value}' not supported for {self.__class__.__name__}. "
                             f"Available: {self._regions}")

        current = getattr(self, "_region", None)
        if current is not None and current != resolved:
            self._data = None
            if hasattr(self, "log"):
                self.log.info(f"Region changed to {resolved}. Cache reset.")

        self._region = resolved
    
    @classmethod
    def list_regions(cls) -> list[str]:
        """List available regions without instantiation."""
        if isinstance(cls._regions, property):
            # reach into the fget of the property
            return cls._regions.fget(cls) 
        return cls._regions




# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import io
import pyarrow as pa
import pyarrow.csv as pv
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.date_utils import offset_period_eom


class MispricingFactors(FactorModel):
    """Mispricing Factors of Stambaugh & Yuan (2016).

    Download and process the Mispricing factor data of Stambaugh-Yuan
    (2016). Data from 1963 to 2016. The SMB factor is returned as SMB_SY.

    Args:
        frequency(str, optional): 'm' (monthly), 'd' (daily)
        start_date (str, optional): The start date YYYY-MM-DD.
        end_date (str, optional): The end date YYYY-MM-DD.
        output_file (str, optional): Optional file path to save to file. 
            Supports csv, pkl.
        classic (bool, optional): returns the classic 4-factor q-factor 
            model. Default: False.
        cache_ttl (int, optional): Cached download time-to-live in 
            seconds (default: 86400).

    References:
        R. F. Stambaugh and Y. Yuan (2016). Mispricing Factors. The Review 
        of Financial Studies, vol. 30, no. 4, pp. 1270–1315.

    """
    @property
    def _frequencies(self) -> list[str]: return ["d", "m"]
    
    @property   # decimal. d/m: MKTRF=6,[FACTORS]=10, RF=5.
    def _precision(self) -> int: return 10

    @property
    def schema(self) -> pa.Schema:
        date_col = "DATE" if self.frequency == "d" else "YYYYMM"
        return pa.schema([
            (date_col, pa.string()),
            ('MKTRF', pa.float64()),
            ('SMB', pa.float64()),
            ('MGMT', pa.float64()),
            ('PERF', pa.float64()),
            ('RF', pa.float64()),
        ])


    def _get_url(self) -> str:
        base_url = "https://finance.wharton.upenn.edu/~stambaug"
        file_name = "M4d" if self.frequency == "d" else "M4"
        return f"{base_url}/{file_name}.csv"


    def _read(self, data: bytes) -> pa.Table:
        """Reads the Mispricing factors CSV."""
        try:
            reader = pv.open_csv(
                io.BytesIO(data),
                read_options=pv.ReadOptions(block_size=1024*1024*2),
                parse_options=pv.ParseOptions(delimiter=','),
                convert_options=pv.ConvertOptions(
                    column_types=self.schema,
                    include_columns=self.schema.names,
                ),
            )
            table = pa.Table.from_batches(reader)

            rename_map = {
                self.schema.names[0]: 'date', 
                'MKTRF': 'Mkt-RF', 
                'SMB': 'SMB_SY',
            }
            renames = [rename_map.get(name, name) for name in table.column_names]

            table = table.rename_columns(renames)
            table = offset_period_eom(table, self.frequency)

            return table.combine_chunks()

        except (pa.ArrowIOError, pa.ArrowInvalid) as e:
            msg = f"{self.__class__.__name__}: reading failed: {e}"
            self.log.error(msg)
            raise ValueError(msg) from e



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import logging
import sys
import textwrap
from importlib.metadata import PackageNotFoundError, version


def _get_version():
    """Avoids importing __init__ for the ver. no."""
    try:
        return version("getfactormodels")
    except PackageNotFoundError:
        return "unknown"


def _cli_list_regions():
    """Helper to display regions and exit."""
    from getfactormodels.models.aqr_models import _AQRModel
    from getfactormodels.models.fama_french import FamaFrenchFactors
    
    print(f"\nFAMA-FRENCH MODELS:\n  {textwrap.fill(', '.join(FamaFrenchFactors.list_regions()), width=70)}")
    print(f"\nAQR MODELS:\n  {textwrap.fill(', '.join(_AQRModel.list_regions()), width=70)}")
    print("\n  Note: accepts aliases 'us', 'jpn', 'uk', and 'ger'.")
    sys.exit(0)


def parse_args() -> argparse.Namespace:
    """CLI arg parser for getfactormodels."""
    parser = argparse.ArgumentParser(
        prog='getfactormodels',
        description='Download datasets for various factor models.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='''Example usage:
    getfactormodels --model ff3 --frequency m
    getfactormodels -m liquidity -f m --start 2000-01-01 --end 2009-12-31
    getfactormodels -m 5 -f m -e 2009-12-31 --extract SMB RF -o '~/file.csv'
    getfactormodels -m Carhart -f d -s 2000-01-01 -x MOM -o file
    getfactormodels -m hml_devil --region jpn
        ''', 
    )

    parser.add_argument('-v', '--version', action='version', version=f'getfactormodels {_get_version()}')
    parser.add_argument('-q', '--quiet', action='store_true', help='Suppress output to console.')

    parser.add_argument('-m', '--model', 
                        nargs="+", # new: multiple models. 
                        metavar="MODEL", help="the model to use, e.g., 'ff3', 'q'."
                        " Accepts 3, 4, 5, 6 in place of ff3, carhart, ff5, ff6.")

    parser.add_argument('-f', '--frequency', type=str, default='m',
                        choices=['d', 'w', 'w2w', 'm', 'q', 'y'], metavar="FREQ",
                        help="Data frequency (default: 'm'). Note: 'w2w' (Wed-to-Wed) is "
                        "only available for q-factors.")

    parser.add_argument('-s', '--start', required=False, metavar="YYYY-MM-DD",
                        help='the start date.')

    parser.add_argument('-e', '--end', required=False, metavar="YYYY-MM-DD",
                        help='the end date.')

    parser.add_argument('-o', '--output', type=str, required=False, default=None, metavar="PATH",
                        help='filename/filepath to save the data to.')

    parser.add_argument('-d', '--drop', nargs='+', metavar='FACTOR', 
                        help="drop specific factor(s) from a model.")

    parser.add_argument('-x', '--extract', nargs='+', metavar="FACTOR",
                        help='extract specific factor(s) from a model.')

    parser.add_argument('-r', '--region', dest='region', metavar='REGION',
                        help="Region or country code for AQR/FF models. "
                        "Use `--list-regions` to see all valid regions.")
    parser.add_argument('--list-regions', action='store_true', help="show all supported regions and exit")
    parser.add_argument('--verbose', action='store_true', help="verbose output (set log to debug)")
    
    #parser.add_argument('--industry-portfolios', type=int, action=???, 
    #                    help="Get Fama-French returns by industry. One of: 5, 10, 12, 20, 30, 38, 49") 
    
    args = parser.parse_args()

    if args.verbose:
        logging.getLogger("getfactormodels").setLevel(logging.DEBUG)
    else:
        logging.getLogger("getfactormodels").setLevel(logging.WARNING)

    return args



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
import os
import sys
from pathlib import Path
import pyarrow.csv as pv
from getfactormodels import models as factor_models
from getfactormodels.models.base import FactorModel, RegionMixin
from getfactormodels.utils.cli import parse_args
from getfactormodels.utils.utils import _generate_filename, _get_model_key

log = logging.getLogger("getfactormodels")

def get_factors(model: str | int | list[str | int] = 3, **kwargs) -> FactorModel: #Self
    """Get and process factor model data.

    This function initializes a specific FactorModel subclass based on the 
    requested model param.

    Args:
        model (str | int): the the name of the factor model. Default: 3
            - Fama-French: '3', '5', '6' ('ff3', 'famafrench3')
            - Carhart: '4', 'carhart', 'car', 'ff4'.
            - q-Factors: 'q', 'q5', or 'qclassic', 'q4'
            - HML Devil: 'hmld' 
            - Betting Against Beta: 'bab'
            - Quality Minus Junk: 'qmj'
            - Pastor-Stambaugh Liquidity: 'liq', 'liquidity'
            - Stambaugh-Yuan Mispricing: 'misp', 'mispricing'
            - Intermediary Capital Risk: 'icr'
            - Daniel-Hirshleifer-Sun Behavioral Factors: 'dhs'
            - Barillas Shanken 6-Factors: 'bs', 'bs6'
        frequency (str): data frequency. Availability varies by model.
            - One of: 'd' 'w' 'w2w' 'm' 'q' or 'y'. Default: 'm'.
        start_date (str): optional start date, YYYY-MM-DD.
        end_date (str): optional end date, YYYY-MM-DD.
        output_file (str): optional path or string to save data to disk.
        **kwargs: model specific parameters:
            - region (str): Geographic region (e.g., 'US', 'Developed')
            - classic (bool): Use the 4-factor version of the Q-model.

    Returns:
        FactorModel: a container object with the requested data. 
            Provides the data as a pa.Table via `.data`, or DataFrame 
            via `.to_pandas()` and `.to_polars()`. Supports the 
            DataFrame Interchange Protocol.

    Raises:
        ValueError: the model key is unrecognized or date is invalid.
        RuntimeError: the data download failed after retries.

    Example:
        >>> m = get_factors(3, start_date="2020-01", frequency='m')
        >>> model.extract(['SMB', 'HML'])
        >>> df = model.to_polars()
    """
    if isinstance(model, list):
        if len(model) == 1:
            model = model[0]  # Extract single item from list
        else:
            from getfactormodels.models.base import ModelCollection
            return ModelCollection(model_keys=model, **kwargs)

    # don't pop, so can check later, or let class handle it.
    region = kwargs.get('region', 'usa')
    model_key = _get_model_key(model)

    model_class_map = {
        "3": "FamaFrenchFactors",
        "5": "FamaFrenchFactors",
        "6": "FamaFrenchFactors",
        "4": "CarhartFactors",
        "Qclassic": "QFactors",
        "HighIncomeCCAPM": "HighIncomeCCAPM",
        "ConditionalCAPM": "ConditionalCAPM",
    }

    class_name = model_class_map.get(model_key, f"{model_key}Factors")
    factor_class = getattr(factor_models, class_name, None)

    if not factor_class:
        raise ValueError(f"Model '{model}' not recognized.")

    if model_key in ("3", "5", "6"):
        kwargs["model"] = model_key
    
    if model_key == "Qclassic":
        kwargs["classic"] = True

    
    if issubclass(factor_class, RegionMixin):
        kwargs['region'] = region
    elif region:
        log.warning(f"  '{class_name}' does not support regions. Ignoring: region '{region}'")
    
    return factor_class(**kwargs)


def main():
    args = parse_args()

    if args.list_regions:
        from getfactormodels.utils.cli import _cli_list_regions
        _cli_list_regions()
        sys.exit(0)

    if not args.model:
        print("Error: The -m/--model argument is required.", file=sys.stderr)
        sys.exit(1)

    try:
        model_obj = get_factors(
            model=args.model,
            frequency=args.frequency,
            start_date=args.start,
            end_date=args.end,
            region=args.region, 
        )

        model_obj.load()
   
        if not len(model_obj):
            log.error("No data returned.")
            sys.exit(1)

    except (ValueError, RuntimeError) as e:
        print(f"ERROR: {e}", file=sys.stderr)
        sys.exit(1)

    if args.extract:
        model_obj.extract(args.extract)
    elif args.drop:
        model_obj.drop(args.drop)

    if args.output:
        model_obj.to_file(args.output)

        if not args.quiet:
            actual_path = Path(args.output).expanduser()
            if actual_path.is_dir():
                actual_path = actual_path / _generate_filename(model_obj)

            print(f"Data saved to: {actual_path.resolve()}", file=sys.stderr)

    # fix: '%%bash' commands are run in a subprocess (output was entire table)
    nb_env = 'ipykernel' in sys.modules or 'JPY_PARENT_PID' in os.environ

    if not sys.stdout.isatty() and not nb_env: # piped
        # model_obj.data's been filtered. Write csv stream of it:
        pv.write_csv(model_obj.data, sys.stdout.buffer)

    else:
        # we're interactive/IPython: print preview of table to stderr. 
        # uses the model_obj's __str__ (which prints the Table preview)
        if not args.quiet:
            sys.stderr.write(f"{str(model_obj)}\n")

if __name__ == "__main__":
    main()



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import pyarrow as pa
import pyarrow.compute as pc
from getfactormodels.models.base import CompositeModel, RegionMixin
from getfactormodels.utils.utils import read_from_fred


class ConditionalCAPM(CompositeModel, RegionMixin):
    """Conditional CAPM (CCAPM) of Jaganathan & Wang (1996).

    An implementations of the 'Premium-Labour', or 'Human Capital'
    CCAPM of Jaganathan and Wang (1996).

    - LBR: Smooth growth in per-capita labor income.
    - PREM: Lagged yield spread (Baa - Aaa) as a conditioning variable.

    References:
        R. Jagannathan and Z. Wang (1996). The conditional CAPM
        and the cross-section of expected returns. The Journal of
        Finance 51: 3–53.

    """
    @property 
    def _regions(self) -> list[str]: return ['usa']
    
    def __init__(self, frequency='m', **kwargs):
        super().__init__(**kwargs)
        self.frequency = frequency.lower()
        self.region = kwargs.get('region', 'usa')

    @property
    def _precision(self) -> int: return 8

    @property
    def _frequencies(self) -> list[str]: return ['m', 'q', 'y']

    @property
    def schema(self) -> pa.Schema:
        return pa.Schema([
            ('date', pa.date32()),
            ('LBR', pa.float64()),
            ('PREM', pa.float64()),
        ])


    def _construct(self, client) -> pa.Table:
        m_series = {
            "POPTHM": "pop",        # Total Population
            "BAA": "baa",
            "AAA": "aaa",
            "A576RC1": "income",    # Compensation of Employees (Monthly), (table 2.2A here)
        }

        # read_from_fred returns offset dates. ('m', not self.frequency)
        t = read_from_fred(m_series, frequency='m', client=client)

        table = self._jw_calc(t)

        table = self._resample(table)

        return table.combine_chunks()


    def _resample(self, table: pa.Table) -> pa.Table:
        """Downsample monthly data to Q or Y using the EOP values."""
        if self.frequency == 'm':
            return table

        months = pc.month(table.column("date"))

        if self.frequency == 'q':
            mask = pc.is_in(months, value_set=pa.array([3, 6, 9, 12]))
        elif self.frequency == 'y':
            mask = pc.equal(months, 12)
        else:
            return table

        return table.filter(mask).combine_chunks()


    def _jw_calc(self, table: pa.Table) -> pa.Table:
        """Calculate JW factors.

        - R_LBR(t): (L(t-1) + L(t-2)) / (L(t-2) + L(t-3)) - 1
        - PREM(t): Lagged spread (Baa - Aaa) at t-1
        """
        # Per Capita Labor Income (L)
        l_pc = pc.divide(table.column("income"), table.column("pop"))
        l_pc = pc.multiply(l_pc, 1000.0)
        
        # Credit spread/default risk (Baa - Aaa), pct to decimal here
        spread = pc.divide(pc.subtract(table.column("baa"), table.column("aaa")), 100.0)

        n_total = table.num_rows
        # 3 lags for LBR calculation (t-1, t-2, t-3). Result starts at index 3.
        n_final = n_total - 3
        if n_final <= 0:
            raise ValueError("Insufficient data to calculate lags.")

        # Calculate R_LBR: slice for the formula
        l_tm1 = l_pc.slice(2, n_final)
        l_tm2 = l_pc.slice(1, n_final)
        l_tm3 = l_pc.slice(0, n_final)

        # (l_tm1 + l_tm2) / (l_tm2 + l_tm3) - 1
        r_lbr = pc.subtract(pc.divide(pc.add(l_tm1, l_tm2), pc.add(l_tm2, l_tm3)), 1.0)

        # PREM is the spread at t-1. Return period t at index 3, t-1 is index 2
        prem_lagged = spread.slice(2, n_final)

        dates = table.column("date").slice(3, n_final)

        return pa.Table.from_arrays(
            [dates, r_lbr, prem_lagged],
            names=["date", "LBR", "PREM"],
        )



### TESTS:
# The calc. Assert to known answer(s) 
# JW freqs: m, q, y last m q4 == 12 == q4 value == yearly value 
# Test not enough data, error  (not enough data returned and range too small, same err) 
# Err: FRED 429 
# docstr test the actual calc 
# 



#!/usr/bin/env python3
# getfactormodels: A Python package to retrieve financial factor model data.
# Copyright (C) 2025 S. Martin <x512@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
from .main import get_factors
from .models import (
    AQR6Factors,
    BABFactors,
    BarillasShankenFactors,
    CarhartFactors,
    ConditionalCAPM,
    DHSFactors,
    FamaFrenchFactors,
    HighIncomeCCAPM,
    HMLDevilFactors,
    ICRFactors,
    LiquidityFactors,
    MispricingFactors,
    QFactors,
    QMJFactors,
    VMEFactors,
    FFIndustryPortfolios,
)

logger = logging.getLogger('getfactormodels')
logger.addHandler(logging.NullHandler())

# TODO: setup logging 
# see: https://docs.python.org/3/howto/logging-cookbook.html#logging-to-multiple-destinations
# or use warnings??

__version__ = "0.0.18"
__all__ = [ 
    "get_factors",
    "BABFactors",
    "BarillasShankenFactors",
    "CarhartFactors",
    "DHSFactors",
    "FamaFrenchFactors",
    "HMLDevilFactors",
    "FFIndustryPortfolios",
    "ICRFactors",
    "LiquidityFactors",
    "MispricingFactors",
    "QFactors",
    "QMJFactors",
    "VMEFactors",
    "AQR6Factors",
    "ConditionalCAPM",
    "HighIncomeCCAPM",
]





# This file part of getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import logging
import time
from pathlib import Path
from diskcache import Cache

logging.basicConfig(level=logging.ERROR)
log = logging.getLogger(__name__)
# TODO: remove/clean up debug messages

class _Cache:
    def __init__(self, cache_dir: str, default_timeout=14400): #4 hour default now with head check.
        cache_path = Path(cache_dir)                           #note: some models going to ttl 1 day,
        self.cache = Cache(str(cache_path))                    #     others until eom. TODO.
        self.default_timeout = default_timeout

        msg = f"CACHE INIT: {cache_path}"
        log.debug(msg)

    @property
    def directory(self) -> str:
        return self.cache.directory

    def get(self, key: str) -> tuple[bytes | None, dict | None]:
        """Retrieves data from cache if valid, otherwise returns None."""
        entry = self.cache.get(key)
        if entry:
            log.debug(f"CACHE HIT: {key[:8]}...")
            if isinstance(entry, dict) and "data" in entry:
                metadata = entry.get("metadata")
                log.debug(f"CACHE DATA: ETag: {metadata.get('etag') if metadata else None}")
                return entry["data"], metadata

        log.debug(f"CACHE MISS: {key[:8]}...")
        return None, None


    def set(self, key: str, data: bytes, metadata: dict | None = None, expire_secs: int | None = None):
        """Saves data (bytes) and etag/last_modified metadata as a single entry."""
        timeout = expire_secs or self.default_timeout
        entry = {
            "data": data,
            "metadata": metadata,
            "expires_at": time.time() + self.default_timeout,
        }
        try:
            self.cache.set(key, entry, expire=timeout)
            log.debug(f"CACHE WRITE: {key[:8]}... (expiry: {timeout}s)")
        except Exception as e:
            log.error(f"Failed to write cache entry: {e}")


    # new: TODO: impl clear cache, optionally by model
    def remove(self, key: str):
        """Removes a specific key from the cache."""
        if self.cache.delete(key):
            log.info(f"Cache entry removed: {key}")
        else:
            log.debug(f"Cache entry not found: {key}")


    def close(self):
        self.cache.close()
        msg = f'CLOSED: {self.cache.__class__.__name__}'
        log.debug(msg)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def clear(self):
        """Removes all data entries from the cache directory."""
        self.cache.clear()

        msg = f"CLEARED: (dir: {self.directory})"
        log.info(msg)



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import io
import zipfile
from typing import Any
import pyarrow as pa
import pyarrow.csv as pv
import pyarrow.compute as pc
from getfactormodels.models.base import FactorModel, RegionMixin
from getfactormodels.utils.arrow_utils import (
    round_to_precision,
    scale_to_decimal,
    print_table_preview,
)
from getfactormodels.utils.date_utils import offset_period_eom


#TODO: break up _read_zip
class FamaFrenchFactors(FactorModel, RegionMixin):
    """Download Fama-French (and Carhart) factor models.

    Downloads the 3-factor (1993), 5-factor (2015), or 6-factor (2018)
    model of Fama & French, or Carhart's (1997) 4-factor model.

    Weekly data only available for the 3-factor model.

    Args:
        model (str | int): model to return. '3' '4' '5' '6' (default: '3')
        frequency (str, optional): frequency of the data. 'd' 'm' 'y' 'w'
            (default: 'm').
        start_date (str, optional): start date of the data. YYYY[-MM-DD].
        end_date (str, optional): end date of the data. YYYY[-MM-DD].
        cache_ttl (int, optional`): cached download time-to-live in seconds 
            (default: 86400).
        region (str, optional): return an international/emerging market 
            model.

    References:
    - E. F. Fama and K. R., 1993. French. Common risk factors in the returns 
      on stocks and bonds. Journal of Financial Economics, vol. 33, no. 1,
      pp. 3–56.
    - E. F. Fama and K. R. French, 2015. A five-factor asset pricing model. 
      Journal of Financial Economics, vol. 116, no. 1, pp. 1–22.
    - E. F. Fama and K. R. French, 2018. Choosing factors. Journal of 
      Financial Economics, vol. 128, no. 2, pp. 234–252, 2018.

    """
    # TODO: NaNs in FamaFrench models! check
    #Note: "-0.9999" should be NaNs! [TODO: FIXME]
    @property
    def _frequencies(self) -> list[str]: return ['d', 'w', 'm', 'y']
    
    @property 
    def _precision(self) -> int: return 6

    @property 
    def _regions(self) -> list[str]:
        return [
            'us', 'emerging', 'developed', 'ex-us', 'europe',
            'japan', 'asia-pacific-ex-japan', 'north-america', 
        ]

    def __init__(self, 
                 frequency: str = 'm', 
                 model: int | str = '3', 
                 region: str | None = 'us', 
                 **kwargs: Any) -> None:
        """Initialize a Fama-French factor model."""
        self.model = str(model)
        super().__init__(frequency=frequency, model=model, **kwargs)
        
        self.region = region

        self._validate_ff_input()
    

    @property
    def schema(self) -> pa.Schema:
        """Fama-French schema for specific model/frequency/region."""
        cols = [("date", pa.string()), 
                ("Mkt-RF", pa.float64()), 
                ("SMB", pa.float64()), 
                ("HML", pa.float64())]
        
        # 5 and 6-factor models: RMW and CMA
        if self.model in ["5", "6"]:
            cols += [("RMW", pa.float64()), 
                     ("CMA", pa.float64())]

        # mom models (4/6): dev/emerging: WML. US 6 factor: UMD. US 4: MOM.
        if self.model in ["4", "6"]:
            is_intl = self.region not in ["us", None]
            mom_name = "WML" if is_intl else ("UMD" if self.model == "6" else "MOM")
            cols.append((mom_name, pa.float64()))
        cols.append(("RF", pa.float64()))
        
        return pa.schema(cols)

    @property
    def _mom_schema(self) -> pa.Schema:
        """Private helper: schema for momentum files. 'val' is placeholder."""
        return pa.schema([("date", pa.string()), 
                          ("val", pa.float64())])
    @property
    def model(self) -> str:
        return self._model
    @model.setter
    def model(self, value: int | str):
        val = str(value)
        if val not in ["3", "4", "5", "6"]:
            raise ValueError(f"Invalid model '{val}'. Options: 3, 4, 5, 6")
        
        if hasattr(self, '_model') and val != self._model:
            self.log.info(f"Model changed from {self._model} to {val}")
            self._data = None
            self._view = None
        self._model = val 

    @property
    def _ff_region_map(self) -> dict[str, str]:
        """Private: maps region input to region URL str"""
        return {
            'us': 'US',
            'emerging': 'Emerging',
            'developed': 'Developed',
            'ex-us': 'Developed_ex_US',
            'europe': 'Europe',
            'japan': 'Japan',
            'asia-pacific-ex-japan': 'Asia_Pacific_ex_Japan',
            'north-america': 'North_America',
        }

    def _validate_ff_input(self) -> None:
        """Validates input for Fama-French models for things getters/setters dont take care of."""
        if self.region != 'us' and self.frequency == 'w':
            raise ValueError(f"Weekly frequency not available for {self.region}.")

        if self.region == "emerging" and self.frequency != 'm':
            raise ValueError("Emerging markets only available in monthly.")

        if self.frequency == 'w' and self.model != "3":
            raise ValueError("Weekly Fama-French data is only available for the 3-factor model.")
 

    def _get_url(self) -> dict[str, str]:
        """Constructs the URLs for downloading Fama-French data."""
        base_url = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp"
        url_list = {}

        ff_url_name = self._ff_region_map.get(self.region)
        
        if ff_url_name == 'Emerging':
            # Emerging: only monthly 5-factors avail
            filename = "Emerging_5_Factors_CSV.zip"
        elif ff_url_name == 'US':
            _model = "F-F_Research_Data_5_Factors_2x3" if self.model in {"5", "6"} else "F-F_Research_Data_Factors"
            freq_map = {'d': '_daily', 'w': '_weekly'}
            suffix = freq_map.get(self.frequency, "")
            filename = f"{_model}{suffix}_CSV.zip"
        else:
            # International regions
            base_model = '3' if self.model in ['3', '4'] else '5'
            daily = '_Daily' if self.frequency == 'd' else ''
            filename = f"{ff_url_name}_{base_model}_Factors{daily}_CSV.zip"
            
        url_list['factors'] = f"{base_url}/{filename}"
        
        # Construct the momentum URL if needed.
        if self.model in ["4", "6"]:
            freq_suffix = "_daily" if self.frequency == 'd' else ""
            if self.region in ['us', None]:
                mom_filename = f"F-F_Momentum_Factor{freq_suffix}_CSV.zip"
            else:
                mom_filename = f"{self.region}_Mom_Factor{freq_suffix.title()}_CSV.zip"
            
            url_list['momentum'] = f"{base_url}/{mom_filename}"

        return url_list


    def _read_zip(self, _data: bytes, use_schema: pa.Schema = None) -> pa.Table:
        """Download and read CSV from zip file."""
        _schema = use_schema if use_schema else self.schema

        with zipfile.ZipFile(io.BytesIO(_data)) as z:
            filename = z.namelist()[0]
            with z.open(filename) as f:
                lines = f.read().decode('utf-8').splitlines()

        is_mom = 'momentum' in filename.lower() or '_mom_' in filename.lower()
        annual_marker = next((i for i, l in enumerate(lines) if "annual factors:" in l.lower()), None)

        if self.frequency == 'y':
            raw_content = lines[annual_marker + 1:] if annual_marker is not None else lines
        else:
            _start = 6 if self.region == 'emerging' else (12 if is_mom else 4)
            _stop = annual_marker if annual_marker else len(lines)
            raw_content = lines[_start:_stop]

        content = [",".join(_schema.names)]
        for line in raw_content:
            clean = line.strip()
            if not clean or "copyright" in clean.lower():
                if "copyright" in clean.lower(): self.copyright = clean
                continue
            
            # Use commas for split, not spaces!
            parts = [p.strip() for p in clean.split(',')]
            
            # only keeps lines that start with a date
            if parts and parts[0].isdigit():
                content.append(",".join(parts))

        return pv.read_csv(
            io.BytesIO("\n".join(content).encode('utf-8')),
            convert_options=pv.ConvertOptions(
                null_values=["-99.99", "-999", "-99.990", "-0.9999"],
                strings_can_be_null=True,
                column_types=_schema,
                include_columns=_schema.names,
            ),
        )   
    

    def _read(self, data: dict[str, bytes]) -> pa.Table:
        main_cols = [f for f in self.schema if f.name not in ["UMD", "MOM", "WML"]]
        
        # Emerging Markets: base file is 5 factors (even for models 3/4)
        if self.region == 'emerging' and self.model in ["3", "4"]:
            main_cols.insert(4, pa.field("RMW", pa.float64()))
            main_cols.insert(5, pa.field("CMA", pa.float64()))

        main_schema = pa.schema(main_cols)
        # data['factors']
        table = self._read_zip(data['factors'], use_schema=main_schema)

        if 'momentum' in data:
            mom_table = self._read_zip(data['momentum'], use_schema=self._mom_schema)
            
            # momentum column name based on schema
            mom_key = next(k for k in ["UMD", "MOM", "WML"] if k in self.schema.names)
            mom_table = mom_table.rename_columns(["date", mom_key])
            
            table = table.join(mom_table, keys="date", join_type="inner").combine_chunks()

        table = table.set_column(0, "date", table.column(0).cast(pa.string()))
        table = offset_period_eom(table, self.frequency)
        table = scale_to_decimal(table)
        
        table = round_to_precision(table, self._precision)

        return table.select(self.schema.names).combine_chunks()



# Idea, testing something
#----------------------------------------------------------------------------
class FFIndustryPortfolios(FactorModel):  #usa only, no RegionMixin.
    _INDUSTRY_PORTFOLIO_SIZES = {5, 10, 12, 17, 30, 38, 48, 49} #TODO
    #TODO: maybe FactorModel to BaseModel or something if Portfolios are using it.
    def __init__(self, 
                 n_portfolios: int = 5, 
                 frequency: str = 'm',
                 weights: str = 'vw',
                 *,
                 dividends: bool = True,
                 **kwargs: Any) -> None:
        self.n_portfolios = n_portfolios
        self.dividends = dividends
        self.weights = weights
        
        super().__init__(frequency=frequency, **kwargs)
    # Validate portfolio size TODO 
    # TODO:__str__
        
    @property
    def _frequencies(self) -> list[str]: return ['d', 'm', 'y']
    
    @property 
    def _precision(self) -> int: return 6

    @property 
    def _regions(self) -> list[str]: return ['us']

    def _get_url(self) -> str:
        """Get a specific industry portfolio URL."""
        base_url = "https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp"
        
        if self.frequency == 'd':
            suffix = "daily_CSV.zip"
        elif not self.dividends:
            suffix = "Wout_Div_CSV.zip"
        else:
            suffix = "CSV.zip"

        filename = f"{self.n_portfolios}_Industry_Portfolios_{suffix}"
        return f"{base_url}/{filename}"

    @property
    def schema(self, column_names: list[str]) -> pa.Schema:
        """Fama-French industry portfolio schema. Uses extracted headers."""
        fields = []
        for name in column_names:
            if name.lower() == 'date':  #note this is applied after rename
                fields.append(pa.field("date", pa.int32()))
            else:
                fields.append(pa.field(name, pa.float64()))
                
        return pa.schema(fields)

    
    def _prepare_file(self, lines: list[str]) -> str:
        """Select correct table and clean whitespace for CSV parsing."""
        w_str = "Value Weighted" if self.weights == 'vw' else "Equal Weighted"
        
        freq_map = {'y': 'Annual', 'd': 'Daily', 'm': 'Monthly'}
        freq_str = freq_map.get(self.frequency, "Monthly")
        
        start_line = f"Average {w_str} Returns -- {freq_str}"
        
        # Find start line 
        idx = next((i for i, ln in enumerate(lines) if start_line in ln), None)
        if idx is None:
            raise ValueError(f"Could not find section: {start_line}")
        
        cleaned = []

        # Extract the headers as lowercase: 'date' + [industries] 
        headers = ("date " + " ".join(lines[idx + 1].split())).lower().replace("-", "_")
        cleaned.append(headers)

        # data begins 2 lines after the headers
        for line in lines[idx + 2:]:
            clean = line.strip()
            
            # stop on a blank or non-numeric line
            if not clean or not clean[0].isdigit():
                break
                
            # 2 spaces for the CSV parser
            cleaned.append("  ".join(clean.split()))
        return "\n".join(cleaned)


    # TODO: Implement in factor class
    def _fix_ff_nulls(self, table: pa.Table) -> pa.Table:
        """Helper to transform Fama-French values representing NaNs to nulls.

        Converts -99.99, -999 values to null.
        """
        # TODO: FIXME: pyarrow.csv should be able to parse the nulls properly. Isn't.
        #   As unlikely as it may be, there are >1.0 values (e.g., Insurance industry,
        #   annual, 1927), so this could convert a legitmate 0.9999 to null.
        ff_nans = [-99.99, -999.0, -0.9999, -9.99, -99.9]
        
        for i, name in enumerate(table.column_names):
            if name.lower() == "date": continue
            col = table.column(name)
            
            # Create a boolean mask which yields True if ff_nans is_in a col of floats.
            mask = pc.is_in(col, value_set=pa.array(ff_nans, type=pa.float64()))
            
            # replace matches with a true null (NaN).
            _null = pa.scalar(None, type=col.type)
            new_col = pc.if_else(mask, _null, col)
            
            table = table.set_column(i, name, new_col)
            
        return table


    def _read(self, data: bytes) -> pa.Table:
        with zipfile.ZipFile(io.BytesIO(data)) as z:  #noqa
            with z.open(z.namelist()[0]) as f:
                lines = f.read().decode('utf-8').splitlines()

        clean_csv = self._prepare_file(lines)

        table = pv.read_csv(
            io.BytesIO(clean_csv.encode('utf-8')),
            parse_options=pv.ParseOptions(),
            convert_options=pv.ConvertOptions(
                null_values=["-99.99", "-999", "-0.9999"],
                column_types={"date": pa.string()}, # str for offset 
            )
        )
        ## FIXME: temp fix. 
        table = self._fix_ff_nulls(table)
        
        table = offset_period_eom(table, self.frequency)
        table = scale_to_decimal(table)
        # NOTE: don't round_to_precision on portfolios
        return table.combine_chunks()

## TODO: LT_rev, ST_rev...



# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
"""AQR Capital Management factor models.

This module provides access to AQR Capital Management's public datasets. These 
models parse multi-sheet Excel workbooks using calamine, and provide regional 
data using the `RegionMixin()`.

Models: HMLDevilFactors, QMJFactors, BABFactors, VMEFactors, AQR6Factors.

"""
import io
from abc import ABC, abstractmethod
from typing import override
import pyarrow as pa
from python_calamine import CalamineWorkbook
from getfactormodels.models.base import (
    CompositeModel,
    FactorModel,
    RegionMixin,
)
from getfactormodels.utils.arrow_utils import (
    round_to_precision,
    select_table_columns,
)
from getfactormodels.utils.date_utils import offset_period_eom
from getfactormodels.utils.http_client import _HttpClient

_AQR_REGIONS = [
    'aus', 'aut', 'bel', 'can', 'che', 'deu', 'dnk', 'esp', 
    'fin', 'fra', 'gbr', 'grc', 'hkg', 'irl', 'isr', 'ita', 
    'jpn', 'nld', 'nor', 'nzl', 'prt', 'sgp', 'swe', 'usa',
    'europe', 'north america', 'pacific', 'global', 'global ex usa',
]

class _AQRModel(FactorModel, RegionMixin):
    """Base class for AQR factor models.

    Notes:
    - These models are slow to download. Daily datasets are 20-30 MB each,
    and the download is rate limited.
    """
    @property
    def _frequencies(self) -> list[str]: return ["d", "m"]
    
    @property
    def _precision(self) -> int: return 8
    
    @property
    def _regions(self) -> list[str]:
        """List of supported AQR countries/regions."""
        return _AQR_REGIONS

    def __init__(self, frequency: str = 'm', cache_ttl: int = 14400, 
                 region: str = 'usa', **kwargs):
        """Initialize a AQR FactorModel.

        Args:
            frequency (str): The frequency of the data. M, D (default: M)
            start_date (str, optional): The start date of the data, YYYY-MM-DD.
            end_date (str, optional): The end date of the data, YYYY-MM-DD.
            region (str, optional): region or country to return data for 
                excluding RF.
            output_file (str, optional): The filepath to save the output data.
            cache_ttl (str): cache time-to-live in seconds.
        """
        self.cache_ttl = cache_ttl
        super().__init__(frequency=frequency, cache_ttl=cache_ttl, **kwargs)
        self.frequency = frequency
        self.region = region


    @override
    def _download(self, client: _HttpClient | None = None) -> bytes:
        url = self._get_url()
        
        if client is not None:
            return client.stream(url, self.cache_ttl, model_name=self.__class__.__name__)
        # direct calls
        with _HttpClient(timeout=15.0) as new_client:
            return new_client.stream(url, self.cache_ttl, model_name=self.__class__.__name__)


    def _aqr_dt_fix(self, d) -> str:
        """Private helper to fix AQR's MM/DD/YYYY format."""
        # dt objects from calamine
        if hasattr(d, 'strftime'): 
            return d.strftime("%Y%m%d")
        
        if isinstance(d, str) and '/' in d:
            m, day, y = d.split('/')
            return f"{y}{m.zfill(2)}{day.zfill(2)}"
        
        return str(d)


    def _get_header(self, rows: list[list]) -> int:
        """Finds row index containing 'DATE'."""
        for i, row in enumerate(rows):
            if row and str(row[0]).strip().upper() == 'DATE':
                return i
        raise ValueError(f"Could not find a header row containing 'DATE' in {self.__class__.__name__}")


    def _process_sheet(self, sheet_name: str, rows: list[list], header_row: int) -> pa.Table:
        """Extracts date and factor values from a single sheet."""
        headers = [str(h).strip().upper() for h in rows[header_row]]
        
        if sheet_name == 'RF':
            col_idx = 1
        else:
            try: # map region to its col in the sheet
                col_idx = headers.index(self.region.upper())
            except ValueError:
                msg = f"Region '{self.region}' not found in sheet '{sheet_name}'. Available: {headers}"
                raise ValueError(msg)

        dates, values = [], []
        for r in rows[header_row + 1:]:
            # Skip empty rows or footers 
            if not r or r[0] is None or r[col_idx] == '': 
                continue
            try:
                dates.append(self._aqr_dt_fix(r[0]))
                values.append(float(r[col_idx]))
            except (ValueError, TypeError):
                continue

        # Maps sheet name to the factor name
        name = self._sheet_map.get(sheet_name, sheet_name)
        col_name = 'RF_AQR' if name == 'RF' else name
        
        return pa.Table.from_pydict({"date": dates, col_name: values})


    def _read(self, data: bytes) -> pa.Table:
        wb = CalamineWorkbook.from_filelike(io.BytesIO(data))
        tables = []

        for sheet_name in self._sheet_map:
            rows = wb.get_sheet_by_name(sheet_name).to_python()
            headers = self._get_header(rows)
            
            t = self._process_sheet(sheet_name, rows, headers)
            tables.append(offset_period_eom(t, self.frequency))
        
        # using left outer join on these models. Uses the factor the 
        #  model's named after. Gets full data for that factor, and only 
        #  filters that factors NaNs. 
        result = tables[0]
        for next_t in tables[1:]:
            result = result.join(next_t, keys='date', join_type='left outer')

        return round_to_precision(result, self._precision).combine_chunks()

    @property
    @abstractmethod
    def _sheet_map(self) -> dict:
        pass


class HMLDevilFactors(_AQRModel):
    """HML Devil factors, Asness and Frazzini (2013).

    References:
        C. Asness and A. Frazzini (2013). The Devil in HML’s Details. 
        The Journal of Portfolio Management, vol. 39, pp. 49–68.
    """
    @property
    def _sheet_map(self):
        return {'HML Devil': 'HML_Devil',
                'MKT': 'Mkt-RF',
                'SMB': 'SMB',
                'UMD': 'UMD',
                'RF': 'RF'}

    def _get_url(self):
        f = 'daily' if self.frequency == 'd' else 'monthly'
        return f'https://www.aqr.com/-/media/AQR/Documents/Insights/Data-Sets/The-Devil-in-HMLs-Details-Factors-{f}.xlsx'

    @property
    def schema(self) -> pa.Schema:
        """Schema for HML Devil."""
        return pa.schema([  
            ('date', pa.string()),
            ('HML Devil', pa.float64()),
            ('MKT', pa.float64()),
            ('SMB', pa.float64()),
            ('UMD', pa.float64()),
            ('RF', pa.float64()),
        ])


class QMJFactors(_AQRModel):
    """Quality Minus Junk (QMJ), Asness, Frazzini & Pedersen (2019).
    
    References:
        C. Asness, A. Frazzini and L. Pedersen, 2019. Quality Minus Junk.
        Review of Accounting Studies 24, no. 1 (2019): 34-112.
    """
    @property
    def schema(self) -> pa.Schema:
        """Schema for QMJ."""
        return pa.schema([  
            ('date', pa.string()),
            ('QMJ Factors', pa.float64()),
            ('MKT', pa.float64()),
            ('SMB', pa.float64()),
            ('HML FF', pa.float64()),
            ('UMD', pa.float64()),
            ('RF', pa.float64()),
        ])

    @property
    def _sheet_map(self):
        return {'QMJ Factors': 'QMJ',
                'MKT': 'Mkt-RF',
                'SMB': 'SMB',
                'HML FF': 'HML',
                'UMD': 'UMD',
                'RF': 'RF'}

    def _get_url(self):
        f = 'Daily' if self.frequency == 'd' else 'Monthly'
        return f'https://www.aqr.com/-/media/AQR/Documents/Insights/Data-Sets/Quality-Minus-Junk-Factors-{f}.xlsx'


class BABFactors(_AQRModel):
    """Betting Against Beta (BAB) factors, Frazzini and Pedersen (2014).

    References:
        Frazzini, A. and Pedersen, L. (2014). Betting against beta. 
        Journal of Financial Economics, 111(1), 1-25.
    """
    @property
    def schema(self) -> pa.Schema:
        return pa.schema([  
            ('date', pa.string()),
            ('BAB Factors', pa.float64()),
            ('MKT', pa.float64()),
            ('SMB', pa.float64()),
            ('HML FF', pa.float64()),
            ('UMD', pa.float64()),
            ('RF', pa.float64()),
        ])
    
    @property
    def _sheet_map(self):
        return {'BAB Factors': 'BAB',
                'MKT': 'Mkt-RF', 
                'SMB': 'SMB',           # SMB_AQR?
                'UMD': 'UMD',
                'HML FF': 'HML',
                'RF': 'RF_AQR'} 

    def _get_url(self):
        f = 'Daily' if self.frequency == 'd' else 'Monthly'
        return f'https://www.aqr.com/-/media/AQR/Documents/Insights/Data-Sets/Betting-Against-Beta-Equity-Factors-{f}.xlsx'


class AQR6Factors(CompositeModel, RegionMixin):
    """AQR 6-Factor Model, Frazzini, Kabiller & Pederson (2018).

    This is the model used 'Buffett's Alpha'. The factors 
    are: Mkt-RF, SMB, HML, UMD, BAB, QMJ.

    References:
        Frazzini, Kabiller & Pedersen (2018). Buffett’s Alpha. 
        Financial Analysts Journal, 74:4, 35-55.
    """
    def __init__(self, frequency: str = 'm', region: str = 'usa', **kwargs):
        super().__init__(frequency=frequency, **kwargs)
        self.region = region 

    @property
    def _regions(self) -> list[str]: return _AQR_REGIONS

    @property
    def _frequencies(self) -> list[str]: return ['m', 'd']

    def _construct(self, client: _HttpClient) -> pa.Table:
        bab_t = BABFactors(frequency=self.frequency, region=self.region).load(client=client)
        qmj_t = QMJFactors(frequency=self.frequency, region=self.region).load(client=client)

        bab = select_table_columns(bab_t.data, ['BAB', 'Mkt-RF', 'SMB'])
        qmj = select_table_columns(qmj_t.data, ['QMJ', 'HML', 'UMD', 'RF_AQR'])  # We use 'HML FF' on BAB and QMJ, renamed HML. AQR's data.

        # Renaming HML to HML_FF: user doesn't know if it's HML Devil otherwise, and RF to RF_AQR. 
        qmj = qmj.rename_columns([
            "HML_FF" if col == "HML" else "RF_AQR" if col == "RF" else col for col in qmj.column_names
        ])

        table = bab.join(qmj, keys='date').combine_chunks()
        return table.select(self.schema.names).cast(self.schema)

    @property
    def schema(self) -> pa.Schema:
        return pa.schema([
            ('date', pa.string()),
            ('Mkt-RF', pa.float64()),
            ('SMB', pa.float64()),
            ('HML_FF', pa.float64()),
            ('UMD', pa.float64()),
            ('BAB', pa.float64()),
            ('QMJ', pa.float64()),
            ('RF_AQR', pa.float64()),
        ])


# NEW/WIP/TESTING/ETC: Value and Momentum Everywhere. --------------------------------- #
# Returns the 'usa' VAL and MOM by defult.
class VMEFactors(_AQRModel):
    """Value and Momentum Everywhere: Asness, Moskowitz, and Pedersen (2013)."""
    @property
    @override     # only AQR model not avail in daily.
    def _frequencies(self) -> list[str]: return ["m"]

    @property
    @override     # different regions to other models.
    def _regions(self) -> list[str]:
        """Regions specific to the VME Excel layout."""
        return [ 
            'everywhere', 'all_equities', 'all_other', 
            'usa', 'uk', 'europe', 'japan',
        ]         # all_equities by default? us?


    @override     # file is different to the other models.
    def _read(self, data: bytes) -> pa.Table:
        wb = CalamineWorkbook.from_filelike(io.BytesIO(data))
        rows = wb.get_sheet_by_name("VME Factors").to_python()
        header_idx = self._get_header(rows)
        _vme_map = {
            'everywhere': 1, 'all_equities': 3, 'all_other': 5,
            'usa': 7, 'uk': 9, 'europe': 11, 'japan': 13,
        }
        
        start_idx = _vme_map.get(self.region.lower(), 7)
        
        dates, val, mom = [], [], []
        # data is from header_idx + 1
        for r in rows[header_idx + 1:]:
            if not r or r[0] is None or r[start_idx] == '':
                continue
            
            dates.append(self._aqr_dt_fix(r[0]))
            val.append(float(r[start_idx]))
            mom.append(float(r[start_idx + 1]))

        table = pa.Table.from_pydict({"date": dates, "VAL": val, "MOM": mom})
        table = offset_period_eom(table, "m")

        return table.cast(self.schema).combine_chunks()


    @property
    def schema(self) -> pa.Schema:
        """Schema for Value and Momentum Everywhere (VME)."""
        return pa.schema([  
            ('date', pa.string()),  # forcing 'DATE' to lower here
            ('VAL', pa.float64()),  # note: actual column names: VALLS_VME_US90
            ('MOM', pa.float64()),  #                            MOMLS_VME_UK90
        ])
    
    @property
    def _sheet_map(self) -> dict: return {'VME Factors': 'VME'}

    def _get_url(self) -> str:
        f = 'Monthly'
        return f'https://www.aqr.com/-/media/AQR/Documents/Insights/Data-Sets/Value-and-Momentum-Everywhere-Factors-{f}.xlsx'



from io import BytesIO
import pyarrow as pa
import pyarrow.compute as pc
from python_calamine import CalamineWorkbook
from getfactormodels.models.base import FactorModel
from getfactormodels.models.fama_french import FamaFrenchFactors
from getfactormodels.utils.arrow_utils import select_table_columns
from getfactormodels.utils.date_utils import offset_period_eom


class HighIncomeCCAPM(FactorModel):
    """Human Capital CAPM (HCAPM), Campbell & Korniotis (2007).

    An implementation of the Human Capital CCAPM (HCAPM), "using the income 
    growth of high income households, rather than aggregate income growth."

    This model replicates the methodlogy of Campbell & Korniotis (2007), 
    using the same dataset from E. Saez website (updated to 2022).

    Only available as annual.

    Factors:
        HCRT1, HCRT5, HCRT10

    References:
        [1] Campbell, S.D. & Korniotis, G.M. 2008. The human capital 
        that matters: expected returns and the income of affluent
        households. Federal Reserve Board, Finance and Economics 
        Discussion Series. 
        [2] E. Saez & T. Piketty, "Income Inequality in the United States, 
        1913-1998" Quarterly Journal of Economics, 118(1), 2003, 1-39.

    """
    @property
    def _frequencies(self) -> list[str]:
        return ['y'] 

    @property
    def schema(self) -> pa.Schema:
        return pa.schema([
            ('date', pa.int32()),
            ('p90', pa.float64()),
            ('p95', pa.float64()),
            ('p99', pa.float64()),
        ])

    def _get_url(self):
        return "https://eml.berkeley.edu/~saez/TabFig2022.xlsx"

    def _read(self, data: bytes) -> pa.Table:
        workbook = CalamineWorkbook.from_filelike(BytesIO(data))
        rows = workbook.get_sheet_by_index(10).to_python()

        clean_rows = [
            r for r in rows[5:] 
            if len(r) >= 4 and all(isinstance(r[i], (int, float)) for i in range(4))
        ]

        raw_table = pa.Table.from_pydict({
            "date": [int(r[0]) for r in clean_rows],
            "p90": [float(r[1]) for r in clean_rows],
            "p95": [float(r[2]) for r in clean_rows],
            "p99": [float(r[3]) for r in clean_rows],
        })

        hcapm_table = self._calc_log_growth(raw_table)
        hcapm_table = offset_period_eom(hcapm_table, self.frequency)

        # Join
        _ff = FamaFrenchFactors(model='3', frequency=self.frequency).load()
        mkt_rf = select_table_columns(_ff.data, ['Mkt-RF', 'RF'])

        return hcapm_table.join(mkt_rf, keys="date", join_type="inner").combine_chunks()

    
    def _calc_log_growth(self, table: pa.Table) -> pa.Table:
        n = table.num_rows
        mapping = {"HCRT10": "p90", "HCRT5": "p95", "HCRT1": "p99"}

        factor_cols = []
        for src_col in mapping.values():
            log_income = pc.ln(table.column(src_col))
            # growth: ln(y_t) - ln(y_{t-1})
            growth = pc.subtract(log_income.slice(1, n-1), log_income.slice(0, n-1))
            factor_cols.append(growth)

        years = table.column("date").slice(1, n-1)

        return pa.Table.from_arrays(
            [years] + factor_cols, 
            names=["date", "HCRT10", "HCRT5", "HCRT1"],
        )




# getfactormodels: https://github.com/x512/getfactormodels
# Copyright (C) 2025-2026 S. Martin <x512@pm.me>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Distributed WITHOUT ANY WARRANTY. See LICENSE for full terms.
import io
import pyarrow as pa
import pyarrow.csv as pv
from getfactormodels.models.base import FactorModel
from getfactormodels.utils.date_utils import (
    offset_period_eom,
    parse_quarterly_dates,
)


class ICRFactors(FactorModel):
    """Download the Intermediary Capital Ratio (He-Kelly-Manela, 2017).

    Monthly and quarterly data start in 1970, daily begins on 1999-05-03.

    References:
        Z. He, B. Kelly, and A. Manela, 2017. Intermediary asset pricing: 
        New evidence from many asset classes. Journal of Financial Economics, 
        vol. 126, no. 1, pp. 1–35.

    ---

    Notes:
        - quarterly data (as of Dec 2025) contains a duplicate entry
        for 2025Q1.
        - NaNs: daily IC_RISK factor doesn't begin until 2008.
        - Precision: Quarterly: 4 decimals before 2013, 18 after.

    Factors
    - IC_RATIO: Intermediary Capital Ratio 
    - IC_RISK: Intermediary Capital Risk Factor 
    - VW_IR: Intermediary Value Weighted Investment Return
    - LEV_SQ: Intermediary Leverage Ratio Squared

    """
    @property
    def _frequencies(self) -> list[str]: return ["d", "m", "q"]

    @property
    def _precision(self) -> int: return 8 if self.frequency == 'd' else 4  # TODO: check 

    @property
    def schema(self) -> pa.Schema:
        date_col = {'d': 'yyyymmdd', 'm': 'yyyymm'}.get(self.frequency, 'yyyyq')

        return pa.schema([
            (date_col, pa.string()),
            ('intermediary_capital_ratio', pa.float64()),
            ('intermediary_capital_risk_factor', pa.float64()),
            ('intermediary_value_weighted_investment_return', pa.float64()),
            ('intermediary_leverage_ratio_squared', pa.float64()),
        ])


    def _get_url(self) -> str:
        _file = {"d": "daily", "m": "monthly", 
                 "q": "quarterly"}.get(self.frequency)

        return f"https://zhiguohe.net/wp-content/uploads/2025/07/He_Kelly_Manela_Factors_{_file}_250627.csv"


    def _read(self, data: bytes):
        try:
            reader = pv.open_csv(
                io.BytesIO(data),
                read_options=pv.ReadOptions(block_size=1024*1024*2),
                convert_options=pv.ConvertOptions(
                    column_types=self.schema,
                    include_columns=self.schema.names,
                    null_values=[".", "NA", "nan", ""],
                ),
            )
            table = pa.Table.from_batches(reader)

            if self.frequency == "q":
                table = parse_quarterly_dates(table)   # TEST: new util for icr, q. 

            table = offset_period_eom(table, self.frequency)
            # TODO: base should report any repeats in date col 
            # TODO: possibly check if last row is a duplicate quarter and drop it 
            output_cols = ["date", "IC_RATIO", "IC_RISK", "VW_IR", "LEV_SQ"]
            return table.rename_columns(output_cols).combine_chunks()

        except (pa.ArrowIOError, pa.ArrowInvalid) as e:
            msg = f"{self.__class__.__name__}: reading failed: {e}"
            self.log.error(msg)
            raise ValueError(msg) from e
